"""time_series.py: RESQML time series class."""

version = '2nd July 2021'

# Nexus is a registered trademark of the Halliburton Company

# At present, no time zone information is handled

import logging
log = logging.getLogger(__name__)
log.debug('resqml_time_series.py version ' + version)

import os
import datetime as dt
import warnings

from resqpy.olio.base import BaseResqpy
import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu
from resqpy.olio.xml_namespaces import curly_namespace as ns



class TimeDuration:
   """A thin wrapper around python's datatime timedelta objects (not a RESQML class)."""

   def __init__(self, days = None, hours = None, minutes = None, seconds = None,
                earlier_timestamp = None, later_timestamp = None):
      """Create a TimeDuration object either from days and seconds or from a pair of timestamps."""

      # for negative durations, earlier_timestamp should be later than later_timestamp
      # or days, hours etc. should typically all be non-positive
      # ie. days = -1, hours = -12 will be a negative one and a half day duration
      # whilst days = -1, hours = 12 will be a negative half day duration
      self.duration = None
      if earlier_timestamp is not None and later_timestamp is not None:
         if earlier_timestamp.endswith('Z'): earlier_timestamp = earlier_timestamp[:-1]   # Trailing Z is not part of iso format
         if later_timestamp.endswith('Z'): later_timestamp = later_timestamp[:-1]
         dt_earlier = dt.datetime.fromisoformat(earlier_timestamp)
         dt_later   = dt.datetime.fromisoformat(later_timestamp)
         self.duration = dt_later - dt_earlier
      else:
         if days is None: days = 0
         if hours is None: hours = 0
         if minutes is None: minutes = 0
         if seconds is None: seconds = 0
         self.duration = dt.timedelta(days = days, hours = hours, minutes = minutes, seconds = seconds)


   def timestamp_after_duration(self, earlier_timestamp):
      """Create a new timestamp from this duration and an earlier timestamp."""

      if earlier_timestamp.endswith('Z'): earlier_timestamp = earlier_timestamp[:-1]
      dt_earlier = dt.datetime.fromisoformat(earlier_timestamp)
      dt_result = dt_earlier + self.duration
      return dt_result.isoformat() + 'Z'


   def timestamp_before_duration(self, later_timestamp):
      """Create a new timestamp from this duration and a later timestamp."""

      if later_timestamp.endswith('Z'): later_timestamp = later_timestamp[:-1]
      dt_later = dt.datetime.fromisoformat(later_timestamp)
      dt_result = dt_later - self.duration
      return dt_result.isoformat() + 'Z'



class TimeSeries(BaseResqpy):
   """Class for RESQML Time Series within RESQML model object."""

   resqml_type = 'TimeSeries'

   def __init__(self, parent_model, uuid = None, time_series_root = None,
                first_timestamp = None, daily = None, monthly = None, quarterly = None, yearly = None,
                title = None, originator = None, extra_metadata = None):
      """Create a TimeSeries object, either from a time series node in parent model, or from given data.

      arguments:
         parent_model (model.Model): the resqpy model to which the time series will belong
         uuid (uuid.UUID, optional): the uuid of a TimeSeries object to be loaded from xml
         time_series_root (xml node, DEPRECATED): the xml root node; use uuid instead
         first_time_stamp (str, optional): the first timestamp (in RESQML format) if not loading from xml;
            this and the remaining arguments are ignored if loading from xml
         daily (non-negative int, optional): the number of one day interval timesteps to start the series
         monthly (non-negative int, optional): the number of 30 day interval timesteps to follow the daily
            timesteps
         quarterly (non-negative int, optional): the number of 90 day interval timesteps to follow the
            monthly timesteps
         yearly (non-negative int, optional): the number of 365 day interval timesteps to follow the
            quarterly timesteps
         title (str, optional): the citation title to use for a new time series;
            ignored if uuid or time_series_root is not None
         originator (str, optional): the name of the person creating the time series, defaults to login id;
            ignored if uuid or time_series_root is not None
         extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the time series;
            ignored if uuid or time_series_root is not None

      returns:
         newly instantiated TimeSeries object

      note:
         a new bespoke time series can be populated by passing the first timestamp here and using the
         add_timestamp() and/or extend_by...() methods

      :meta common:
      """

      self.timestamps = []    # ordered list of timestamp strings in resqml/iso format
      if first_timestamp is not None:
         self.timestamps.append(first_timestamp)  # todo: check format of first_timestamp
         if daily is not None:
            for _ in range(daily): self.extend_by_days(1)
         if monthly is not None:
            for _ in range(monthly): self.extend_by_days(30)
         if quarterly is not None:
            for _ in range(quarterly): self.extend_by_days(90)   # could use 91
         if yearly is not None:
            for _ in range(yearly): self.extend_by_days(365)  # could use 360
      super().__init__(model = parent_model, uuid = uuid, title = title, originator = originator,
                       extra_metadata = extra_metadata, root_node = time_series_root)


   def _load_from_xml(self):
      time_series_root = self.root
      assert time_series_root is not None
      for child in self.time_series_root:
         if rqet.stripped_of_prefix(child.tag) != 'Time': continue
         dt_node = rqet.find_tag(child, 'DateTime')
         if dt_node is None: continue          # could raise an exception here
         self.timestamps.append(dt_node.text)  # todo: trim and check timestamp
         self.timestamps.sort()


   def is_equivalent(self, other_ts, tol_seconds = 1):
      """Returns True if the this timestep series is essentially identical to the other; otherwise False."""

      if other_ts is None: return False
      if self is other_ts: return True
      if bu.matching_uuids(self.uuid, other_ts.uuid): return True
      if self.number_of_timestamps() != other_ts.number_of_timestamps(): return False
      tolerance = TimeDuration(seconds = tol_seconds)
      for t_index in range(self.number_of_timestamps()):
         diff = TimeDuration(earlier_timestamp = self.timestamps[t_index],
                             later_timestamp = other_ts.timestamps[t_index])
         if abs(diff.duration) > tolerance.duration: return False
      return True


   def set_model(self, parent_model):
      """Associate the time series with a resqml model (does not create xml or write hdf5 data)."""
      self.model = parent_model


   def number_of_timestamps(self):
      """Returns the number of timestamps in the series.

      :meta common:
      """
      return len(self.timestamps)


   def timestamp(self, index):
      """Returns an individual timestamp, indexed by its position in the series.

      :meta common:
      """

      # individual timestamp is in iso format with a Z appended, eg: 2019-08-23T14:30:00Z
      if index < 0 or index >= len(self.timestamps): return None
      return self.timestamps[index]


   def iter_timestamps(self):
      """Iterator over timestamps.

      :meta common:
      """
      for t_stamp in self.timestamps:
         yield t_stamp


   def last_timestamp(self):
      """Returns the last timestamp in the series.

      :meta common:
      """
      return self.timestamps[-1] if self.timestamps else None


   def index_for_timestamp(self, timestamp):
      """Returns the index for a given timestamp.

      note:
         this method uses a simplistic string comparison

      :meta common:
      """
      for index in range(len(self.timestamps)):
         if self.timestamps[index] == timestamp: return index
      return None


   def index_for_timestamp_not_later_than(self, timestamp):
      """Returns the index of the latest timestamp that is not later than the specified date.

      :meta common:
      """
      index = len(self.timestamps) - 1
      while (index >= 0) and (self.timestamps[index] > timestamp): index -= 1
      if index < 0: return None
      return index


   def index_for_timestamp_not_earlier_than(self, timestamp):
      """Returns the index of the earliest timestamp that is not earlier than the specified date.

      :meta common:
      """
      index = 0
      while (index < len(self.timestamps)) and (self.timestamps[index] < timestamp): index += 1
      if index >= len(self.timestamps): return None
      return index


   def index_for_timestamp_closest_to(self, timestamp):
      """Returns the index of the timestamp that is closest to the specified date.

      :meta common:
      """
      if not self.timestamps: return None
      before = self.index_for_timestamp_not_later_than(self, timestamp)
      if not before: return 0
      if before == len(self.timestamps) - 1 or self.timestamps[before] == timestamp: return before
      after = before + 1
      early_delta = TimeDuration(earlier_timestamp = self.timestamps[before], later_timestamp = timestamp)
      later_delta = TimeDuration(earlier_timestamp = timestamp, later_timestamp = self.timestamps[after])
      return before if early_delta <= later_delta else after


   def duration_between_timestamps(self, earlier_index, later_index):
      """Returns the duration between a pair of timestamps.

      :meta common:
      """
      if earlier_index < 0 or later_index >= len(self.timestamps) or later_index < earlier_index: return None
      return TimeDuration(earlier_timestamp = self.timestamps[earlier_index],
                          later_timestamp = self.timestamps[later_index])


   def days_between_timestamps(self, earlier_index, later_index):
      """Returns the number of whole days between a pair of timestamps, as an integer."""
      delta = self.duration_between_timestamps(earlier_index, later_index)
      if delta is None: return None
      return delta.duration.days


   def duration_since_start(self, index):
      """Returns the duration between the start of the time series and the indexed timestamp.

      :meta common:
      """
      if index < 0 or index >= len(self.timestamps): return None
      return self.duration_between_timestamps(0, index)


   def days_since_start(self, index):
      """Returns the number of days between the start of the time series and the indexed timestamp."""
      return self.duration_since_start(index).duration.days


   def step_duration(self, index):
      """Returns the duration of the time step between the indexed timestamp and preceeding one.

      :meta common:
      """
      if index < 1 or index >= len(self.timestamps): return None
      return self.duration_between_timestamps(index - 1, index)


   def step_days(self, index):
      """Returns the number of days between the indexed timestamp and preceeding one."""
      delta = self.step_duration(index)
      if delta is None: return None
      return delta.duration.days


   # NB: Following functions modify the time series, which is dangerous if the series is in use by a model
   # Could check for relationships involving the time series and disallow changes if any found?
   def add_timestamp(self, new_timestamp, allow_insertion = False):
      """Inserts a new timestamp into the time series."""
      # todo: check that new_timestamp is in valid format (iso format + 'Z')
      if allow_insertion:
         # NB: This can insert a timestamp anywhere in the series, which will invalidate indices, possibly corrupting model
         index = self.index_for_timestamp_not_later_than(new_timestamp)
         if index is None: index = 0
         else: index += 1
         self.timestamps.insert(new_timestamp, index)
      else:
         last = self.last_timestamp()
         if last is not None: assert(new_timestamp > self.last_timestamp())
         self.timestamps.append(new_timestamp)


   def extend_by_duration(self, duration):
      """Adds a timestamp to the end of the series, at duration beyond the last timestamp."""
      assert(duration.duration.days >= 0)    # duration may not be negative
      assert(len(self.timestamps) > 0)       # there must be something to extend from
      self.timestamps.append(duration.timestamp_after_duration(self.last_timestamp()))


   def extend_by_days(self, days):
      """Adds a timestamp to the end of the series, at a duration of days beyond the last timestamp."""
      duration = TimeDuration(days = days)
      self.extend_by_duration(duration)


   def datetimes(self):
      """Returns a list of python-datetime objects """
      return [dt.datetime.fromisoformat(t.rstrip('Z')) for t in self.timestamps]


   def create_xml(self, add_as_part = True, title = None, originator = None, reuse = True):
      """Create a time series node from a TimeSeries object, optionally add as part.

      arguments:
         add_as_part (boolean, default True): if True, the newly created xml node is added as a part
            in the model
         title (string): used as the citation Title text for the new time series node
         originator (string, optional): the name of the human being who created the time series object;
            default is to use the login name

      returns:
         the newly created time series xml node

      :meta common:
      """

      if reuse and self.try_reuse(): return self.node  # check for reusable (equivalent) object

      ts_node = super().create_xml(add_as_part = False, title = title, originator = originator)

      for index in range(self.number_of_timestamps()):
         time_node = rqet.SubElement(ts_node, ns['resqml2'] + 'Time')
         time_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Timestamp')
         time_node.text = rqet.null_xml_text
         dt_node = rqet.SubElement(time_node, ns['resqml2'] + 'DateTime')
         dt_node.set(ns['xsi'] + 'type', ns['xsd'] + 'dateTime')
         dt_node.text = self.timestamp(index)

      if add_as_part: self.model.add_part('obj_TimeSeries', self.uuid, ts_node)

      return ts_node


   def create_time_index(self, time_index, root = None):
      """Create a time index node, including time series reference, optionally add to root.

      arguments:
         time_index (int): non-negative integer index into the time series, identifying the datetime
            being referenced by the new node
         root (optional): if present, the newly created time index xml node is added
            as a child to this node

      returns:
         the newly created time index xml node

      note:
         a time index node is typically part of a recurrent grid property object; it identifies
         the datetime relevant to the property array (or other data) by indexing into a time series
         object
      """

      assert self.uuid is not None

      t_node = rqet.Element(ns['resqml2'] + 'TimeIndex')
      t_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TimeIndex')
      t_node.text = rqet.null_xml_text

      index_node = rqet.SubElement(t_node, ns['resqml2'] + 'Index')
      index_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
      index_node.text = str(time_index)

      self.model.create_time_series_ref(self.uuid, root = t_node)

      if root is not None: root.append(t_node)
      return t_node


   @property
   def time_series_root(self):
      """DEPRECATED. Alias for root"""
      warnings.warn("Attribute 'time_series_root' is deprecated. Use 'root'", DeprecationWarning)
      return self.root


def selected_time_series(full_series, indices_list, title = None):
   """Returns a new TimeSeries object with timestamps selected from the full series by a list of indices."""

   selected_ts = TimeSeries(full_series.model, title = title)
   selected_ts.timestamps = [full_series.timestamps[i] for i in sorted(indices_list)]
   return selected_ts


def simplified_timestamp(timestamp):
   """Return a more readable version of the timestamp."""

   if timestamp is None: return None
   timestamp = cleaned_timestamp(timestamp)
   if len(timestamp) > 10: timestamp = timestamp[:10] + ' ' + timestamp[11:]
   return timestamp[:19]


def cleaned_timestamp(timestamp):
   """Return a cleaned version of the timestamp."""

   if timestamp is None: return None
   timestamp = str(timestamp)
   if len(timestamp) < 19 or timestamp[11:19] == '00:00:00': return timestamp[:10]
   return timestamp[:10] + 'T' + timestamp[11:19] + 'Z'


def time_series_from_list(timestamp_list, parent_model = None):
   """Create a TimeSeries object from a list of timestamps (model and node set to None)."""

   assert(len(timestamp_list) > 0)
   sorted_timestamps = sorted(timestamp_list)
   time_series = TimeSeries(parent_model = parent_model, first_timestamp = cleaned_timestamp(sorted_timestamps[0]))
   for raw_timestamp in sorted_timestamps[1:]:
      timestamp = cleaned_timestamp(raw_timestamp)
      time_series.add_timestamp(timestamp)
   return time_series


def merge_timeseries_from_uuid(model, timeseries_uuid_iter):
   """Create a TimeSeries object from an iteratable object of existing timeseries UUIDs of timeseries. iterable can be a list, array, or iteratable generator (model must be provided). The new timeseries is sorted in ascending order. Returns the new time series, the new time series uuid, and the list of timeseries objects used to generate the list"""
   
   reverse=False 

   alltimestamps=set({})
   timeserieslist=[]
   for timeseries_uuid in timeseries_uuid_iter:
       timeseriesroot=model.root(uuid=timeseries_uuid)
       assert(rqet.node_type(timeseriesroot) == 'obj_TimeSeries')
       
       singlets=TimeSeries(model, time_series_root=timeseriesroot)
       timeserieslist.append(singlets)
       #alltimestamps.update( set(singlets.timestamps) )
       alltimestamps.update( set(singlets.datetimes()) )
   
   sortedtimestamps=sorted(list(alltimestamps), reverse=reverse)
    
   new_time_series=time_series_from_list(sortedtimestamps, parent_model=model)
   new_time_series_uuid=new_time_series.uuid
   return (new_time_series, new_time_series_uuid, timeserieslist)


def time_series_from_nexus_summary(summary_file, parent_model = None):
   """Create a TimeSeries object based on time steps reported in a nexus summary file (.sum)."""

   if not summary_file: return None
   if not os.path.isfile(summary_file):
      log.warning('Summary file not found: ' + summary_file)
      return None
   try:
      us_date_format = True
      date_format = 'MM/DD/YYYY'
      summary_entries = []      # list of (time step no., cumulative time (days), date (dd/mm/yyyy))
      with open(summary_file, "r") as fp:
         while True:
            line = fp.readline()
            if len(line) == 0: break   # end of file
            words = line.split()
            if len(words) < 5: continue
            if not words[0].isdigit():
               if words[0].upper() == 'NO.' and len(words[4]) == 10:
                  date_format = words[4].upper()
                  us_date_format = (date_format == 'MM/DD/YYYY')
                  if not us_date_format: assert date_format == 'DD/MM/YYYY'
               continue
            ts_number = int(words[0])
            time_in_days = float(words[3])
            date_str = words[4]
            assert 9 <= len(date_str) <= 10, 'invalid date string length in summary file: ' + date_str
            if us_date_format:
                date = dt.date(int(date_str[-4:]), int(date_str[:-8]), int(date_str[-7:-5]))
            else:
                date = dt.date(int(date_str[-4:]), int(date_str[-7:-5]), int(date_str[:-8]))
            summary_entries.append((ts_number, time_in_days, date))
      if len(summary_entries) == 0: return None      # no entries extracted from summary file, could raise error?
      summary_entries.sort()
      if summary_entries[0][0] == 0:   # first entry is for time zero
         tz_date = summary_entries[0][2]
         summary_entries.pop(0)
      else:   # back calculate time zero from first entry
         delta = dt.timedelta(days = summary_entries[0][1])
         tz_date = summary_entries[0][2] - delta
      time_series = TimeSeries(parent_model = parent_model, first_timestamp = tz_date.isoformat() + 'T00:00:00Z')
      last_timestep = 0
      last_cumulative_time = 0.0
      for entry in summary_entries:
         if entry[0] <= last_timestep:
            log.warning('ignoring out of sequence time step in summary file: ' + str(entry[0]))
            continue
         while entry[0] > last_timestep + 1:   # add filler steps
            time_series.extend_by_days(0)
            last_timestep += 1
         if entry[1] < last_cumulative_time:   # should never happen
            time_series.extend_by_days(0)
         else:
            time_series.extend_by_days(entry[1] - last_cumulative_time)
            last_cumulative_time = entry[1]
         last_timestep += 1
      return time_series
   except:
      log.exception('failed to create TimeSeries object from summary file: ' + summary_file)
   return None



import resqpy.olio.uuid as bu
from resqpy.olio.base import BaseResqpy


class DummyObj(BaseResqpy):
    resqml_type = 'DummyResqmlInterpretation'


class ReusableDummyObj(BaseResqpy):
    resqml_type = 'ReusableDummyResqmlInterpretation'

    # a class must have this method to allow meaningful reusability
    def is_equivalent(self, other):
       return True


def test_base_creation(tmp_model):

    # Setup new object
    title = 'Wondermuffin'
    dummy = DummyObj(model=tmp_model, title=title)

    # UUID should exist
    # but root should not exist yet and part name does not exist in the parent model
    assert dummy.uuid is not None
#   assert dummy.part is not None  # this version constructed the part from resqml_type and uuid
    assert dummy.part is None
    assert dummy.root is None

    # After creating XML, root should exist
    dummy.create_xml(add_as_part=True)
    assert dummy.root is not None


def test_base_save_and_load(tmp_model):

    # Create and save a DummyObj
    title = 'feefifofum'
    originator = 'Scruffian'
    metadata = {"balderdash": "pumpernickel"}
    dummy1 = DummyObj(model=tmp_model, title=title, originator=originator)
    dummy1.extra_metadata = metadata

    # Save to XML
    dummy1.create_xml(add_as_part=True)

    # Load a new object
    dummy2 = DummyObj(model=tmp_model, uuid=dummy1.uuid)

    # Properties should match
    assert bu.matching_uuids(dummy2.uuid, dummy1.uuid)
    assert dummy2.root is not None
    assert dummy2.title == title
    assert dummy2.originator == originator
    assert dummy2.extra_metadata == metadata


def test_base_comparison(tmp_model):

    # Setup new object
    dummy1 = DummyObj(model=tmp_model)
    dummy2 = DummyObj(model=tmp_model, uuid=dummy1.uuid)  # Same UUID
    dummy3 = DummyObj(model=tmp_model, title='hello')     # Different UUID

    # Comparison should work by UUID
    assert dummy1 == dummy2
    assert dummy1 != dummy3

    
def test_base_repr(tmp_model):

    dummy = DummyObj(model=tmp_model)

    # Check repr method produces a string
    rep = repr(dummy)
    assert isinstance(rep, str)
    assert len(rep) > 0

    # Check HTML can be generated
    html = dummy._repr_html_()
    assert len(html) > 0


def test_base_reuse_duplicate(tmp_model):

    # Create object and save
    dummy_1 = ReusableDummyObj(model=tmp_model)
    assert not dummy_1.try_reuse()
    dummy_1.create_xml(add_as_part=True)

    # Should be present in model
    uuids = tmp_model.uuids(obj_type=ReusableDummyObj.resqml_type)
    assert len(uuids) == 1
    assert dummy_1.try_reuse()  # after a part has had xml created and added to model, it is reusable

    # Create duplicate object
    dummy_2 = ReusableDummyObj(model=tmp_model)
    assert not bu.matching_uuids(dummy_1.uuid, dummy_2.uuid)
    assert dummy_2.try_reuse()  # should have matched dummy_1
    assert bu.matching_uuids(dummy_1.uuid, dummy_2.uuid)

    # Only one 'RESQML' object should exist in model, despite us having two resqpy objects
    uuids = tmp_model.uuids(obj_type=ReusableDummyObj.resqml_type)
    assert len(uuids) == 1
    assert bu.matching_uuids(uuids[0], dummy_2.uuid)



from numpy.lib.arraysetops import isin
import pytest
import os

import resqpy.model as rq
import resqpy.crs as rqc
import resqpy.well as rqw
import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu

def test_model(tmp_path):

   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   assert model is not None
   crs = rqc.Crs(model)
   crs_root = crs.create_xml()
   model.store_epc()
   assert os.path.exists(epc)
   md_datum_1 = rqw.MdDatum(model, location = (0.0, 0.0, -50.0), crs_root = crs_root)
   md_datum_1.create_xml(title = 'Datum 1')
   md_datum_2 = rqw.MdDatum(model, location = (3.0, 0.0, -50.0), crs_root = crs_root)
   md_datum_2.create_xml(title = 'Datum 2')
   assert len(model.uuids(obj_type = 'MdDatum')) == 2
   model.store_epc()

   model = rq.Model(epc)
   assert model is not None
   assert len(model.uuids(obj_type = 'MdDatum')) == 2
   datum_part_1 = model.part(obj_type = 'MdDatum', title = '1', title_mode = 'ends')
   datum_part_2 = model.part(obj_type = 'MdDatum', title = '2', title_mode = 'ends')
   assert datum_part_1 is not None and datum_part_2 is not None and datum_part_1 != datum_part_2
   datum_uuid_1 = rqet.uuid_in_part_name(datum_part_1)
   datum_uuid_2 = rqet.uuid_in_part_name(datum_part_2)
   assert not bu.matching_uuids(datum_uuid_1, datum_uuid_2)
   p1 = model.uuid_part_dict[bu.uuid_as_int(datum_uuid_1)]
   p2 = model.uuid_part_dict[bu.uuid_as_int(datum_uuid_2)]
   assert p1 == datum_part_1 and p2 == datum_part_2


def test_model_iterators(example_model_with_well):
   
   model, well_interp, datum, traj = example_model_with_well

   w = next(model.iter_wellbore_interpretations())
   d = next(model.iter_md_datums())
   t = next(model.iter_trajectories())

   assert bu.matching_uuids(w.uuid, well_interp.uuid)
   assert bu.matching_uuids(d.uuid, datum.uuid)
   assert bu.matching_uuids(t.uuid, traj.uuid)
   assert w == well_interp
   assert d == datum
   assert t == traj


def test_model_iterate_objects(example_model_with_well):

   from resqpy.organize import WellboreFeature, WellboreInterpretation

   model, well_interp, _, _ = example_model_with_well
   
   # Try iterating over wellbore features

   wells = model.iter_objs(WellboreFeature)
   wells = list(wells)
   w = wells[0]
   assert len(wells) == 1
   assert isinstance(wells[0], WellboreFeature)
   assert wells[0].title == "well A"
   
   # Try iterating over wellbore interpretations

   interps = model.iter_objs(WellboreInterpretation)
   interps = list(interps)
   assert len(interps) == 1
   assert isinstance(interps[0], WellboreInterpretation)
   assert interps[0] == well_interp


def test_model_iter_crs(example_model_and_crs):
   
   model, crs_1 = example_model_and_crs

   crs_list = list(model.iter_crs())
   assert len(crs_list) == 1

   crs_2 = crs_list[0]
   assert crs_2 == crs_1

def test_model_iter_crs_empty(tmp_model):

   # Should raise an exception if no CRS exists
   with pytest.raises(StopIteration):
      next(tmp_model.iter_crs())


def test_model_as_graph(example_model_with_well):

   model, well_interp, datum, traj = example_model_with_well
   crs = next(model.iter_crs())

   nodes, edges = model.as_graph()

   # Check nodes
   for obj in [well_interp, datum, traj, crs]:
      assert str(obj.uuid) in nodes.keys()
      if hasattr(obj, 'title'):  # TODO: remove this when all objects have this attribute
         assert obj.title == nodes[str(obj.uuid)]['title']

   # Check edges
   assert frozenset([str(traj.uuid), str(datum.uuid)]) in edges
   assert frozenset([str(traj.uuid), str(crs.uuid)]) in edges
   assert frozenset([str(traj.uuid), str(well_interp.uuid)]) in edges
   assert frozenset([str(datum.uuid), str(traj.uuid)]) in edges

   # Test uuid subset
   nodes, edges = model.as_graph(uuids_subset=[datum.uuid])
   assert len(nodes.keys()) == 1
   assert len(edges) == 0



"""tomography.py: high level functions to support 2D mapping and generation of cross sectional images."""

# todo: cross section functions

version = '29th April 2021'

import logging
log = logging.getLogger(__name__)
log.debug('tomography.py version ' + version)

import os
import numpy as np
from PIL import Image, ImageDraw

import resqpy.surface as rqs
import resqpy.property as rqp
import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu


def make_pixel_map(grid, width, height, origin = None, dx = None, dy = None, border = 0.0,
                   k0 = 0, vertical_ref = 'top', trim = True):
   """Generate a regular mesh and related pixel map property, writing to hdf5 and adding as parts.

   arguments:
      grid (resqpy.grid.Grid object): the (usually irregular) source grid which is to be mapped
      width (int): the width of the pixel rectangle (number of pixels)
      height (int): the height of the pixel rectangle (number of pixels)
      origin (float pair, optional): x, y of south west corner of area covered by pixel rectangle,
         in local crs of grid; if None then the minimum point of the grid's box is used and border
         applied
      dx (float, optional): the size (west to east) of a pixel, in locel crs of grid; if dx and dy
         are both None, they are computed from the grid's box, border, width and height
      dy (float, optional): the size (south to north) of a pixel, in locel crs; if only one of dx
         or dy is given, the other is set the same; if both None they are computed (see dx)
      border (float, default 0.0): a border width, in local crs of grid, applied around bounding box
         of grid when computing origin and/or dx & dy; ignored if those values are all supplied
      k0 (int, default 0): the layer to create a 2D pixel map for
      vertical_ref (string, default 'top'): 'top' or 'base'
      trim (boolean, default True): if True, the width or height may be reduced so that the resulting
         pixel map has an equal border on all 4 sides

   returns:
      resqpy.surface.Mesh, resqpy.property.GridPropertyCollection

   note:
      this function does not re-write the epc but it does write new pixel map data to hdf5 and create xml
   """

   log.debug(f'making pixel map with width, height: {width}, {height}')

   assert grid is not None
   assert width > 0 and height > 0
   assert border >= 0.0

   model = grid.model
   assert model is not None

   if origin is None or (dx is None and dy is None):
      grid_xyz_box = grid.xyz_box(lazy = False, local = True)
      if origin is None: origin = tuple(grid_xyz_box[0, :2] - border)
      if dx is None and dy is None:
         max_x, max_y = grid_xyz_box[1, :2] + border
         dx = (max_x - origin[0]) / float(width)
         dy = (max_y - origin[1]) / float(height)
         if dx > dy:
            dy = dx
            if trim: height = int((max_y - origin[1]) / dy)
         else:
            dx = dy
            if trim: width = int((max_x - origin[0]) / dx)
      elif dx is None:
         dx = dy
      elif dy is None:
         dy = dx
   assert len(origin) == 2
   origin_xyz = tuple(list(origin) + [0.0])
   assert dx > 0.0 and dy > 0.0

   p_map = grid.pixel_maps(origin, width, height, dx, dy = dy, k0 = k0, vertical_ref = vertical_ref)

   source_str = 'pixel map for grid ' + rqet.citation_title_for_node(grid.root)

   p_mesh = rqs.Mesh(parent_model = model, nj = height + 1, ni = width + 1,
                     origin = origin_xyz, dxyz_dij = np.array([[dx, 0.0, 0.0], [0.0, dy, 0.0]]),
                     crs_uuid = grid.crs_uuid)

   p_mesh.create_xml(title = 'pixel map frame')

   property_collection = rqp.PropertyCollection()
   property_collection.set_support(support = p_mesh)

   p_uuid = property_collection.add_cached_array_to_imported_list(
      p_map,
      source_str,
      keyword = 'pixel map',
      discrete = True, null_value = -1,
      property_kind = 'index', facet_type = 'what', facet = 'pixel map',
      indexable_element = 'cells', count = 2)

   property_collection.write_hdf5_for_imported_list()
   property_collection.create_xml_for_imported_list_and_add_parts_to_model()

   # todo: add extra metadata to p_map property holding uuid of grid

   p_node = model.root_for_uuid(p_uuid)

   model.create_reciprocal_relationship(p_node, 'destinationObject', grid.root, 'sourceObject')

   log.debug('pixel map made')

   return p_mesh, property_collection


def find_pixel_maps(grid, min_width = None, max_width = None, min_height = None, max_height = None):
   """Returns a property collection of existing pixel maps in model for mapping grid, optionally filtered by resolution.

   arguments:
      grid (resqpy.grid.Grid object): the (usually irregular) source grid which is to be mapped
      min_width (int, optional): if present, only pixel maps of at least this width are included
      max_width (int, optional): if present, only pixel maps of at most this width are included
      min_height (int, optional): if present, only pixel maps of at least this height are included
      max_height (int, optional): if present, only pixel maps of at most this height are included

   returns:
      resqpy.property.PropertyCollection containing the existing pixel maps for mapping the grid object
   """

   log.debug('finding pixel maps')

   model = grid.model

   filtering = (min_width or max_width or min_height or max_height)

   # pixel maps have a secondary relationship with the grid for mapping
   property_parts_list = model.parts_list_of_type(type_of_interest = 'obj_DiscreteProperty')
   log.debug('number of discrete properties: ' + str(len(property_parts_list)))
   property_parts_list = model.parts_list_filtered_by_related_uuid(property_parts_list, grid.uuid)
   log.debug('number of discrete properties related to grid: ' + str(len(property_parts_list)))
   # property parts list will contain pixel maps and native discrete properties for grid

   # following property collections may contain parts with various supporting representations
   pm_collection = rqp.PropertyCollection()
   pm_collection.model = model
   pm_collection.add_parts_list_to_dict(property_parts_list)
   log.debug('initial property collection size: ' + str(pm_collection.number_of_parts()))

   pm_collection = rqp.selective_version_of_collection(pm_collection,
                                                       count = 2,
                                                       property_kind = 'index',
                                                       facet_type = 'what',
                                                       facet = 'pixel map',
                                                       indexable = 'cells')
   log.debug('selective property collection size: ' + str(pm_collection.number_of_parts()))

   if filtering:
      log.debug('filtering')
      pixel_map_mesh_uuid = None
      filtered_parts_list = []
      for part in pm_collection.parts():
         support_uuid = model.supporting_representation_for_part(part)
         if support_uuid is None: continue
         if pixel_map_mesh_uuid is None or not bu.matching_uuids(support_uuid, pixel_map_mesh_uuid):
            pixel_map_mesh_root = model.root_for_uuid(support_uuid)
            pixel_map_extent_ji = np.ones(2, dtype = 'int')
            pixel_map_extent_ji[0] = rqet.find_nested_tags_int(pixel_map_mesh_root, ['Grid2dPatch', 'SlowestAxisCount']) - 1
            pixel_map_extent_ji[1] = rqet.find_nested_tags_int(pixel_map_mesh_root, ['Grid2dPatch', 'FastestAxisCount']) - 1
            pixel_map_mesh_uuid = support_uuid
         if min_width is not None and (pixel_map_extent_ji[1] < min_width): continue
         if max_width is not None and (pixel_map_extent_ji[1] > max_width): continue
         if min_height is not None and (pixel_map_extent_ji[0] < min_height): continue
         if max_height is not None and (pixel_map_extent_ji[0] > max_height): continue
         filtered_parts_list.append(part)
      pm_collection = rqp.PropertyCollection()
      pm_collection.model = model
      pm_collection.add_parts_list_to_dict(filtered_parts_list)

   if pm_collection.number_of_parts() == 0: pm_collection = None

   if pm_collection:
      log.debug('found ' + str(pm_collection.number_of_parts()) + ' pixel maps')
   else:
      log.debug('no pixel maps found')

   return pm_collection


def find_or_make_pixel_map(grid, width, height, exact_resolution = False,
                           min_width = None, max_width = None, min_height = None, max_height = None,
                           origin = None, dx = None, dy = None, border = 0.0,
                           k0 = 0, vertical_ref = 'top', trim = True):
   """Looks for pixel map(s) for grid and returns if found; otherwise makes one.

   arguments:
      width (int): the width of the pixel rectangle (number of pixels)
      height (int): the height of the pixel rectangle (number of pixels)
      exact_resolution (bool, default False): if True, equivalent to setting the following min & max
         arguments equal to width and to height, forcing an exact match to the given resolution
      min_width (int, optional): if present, only pixel maps of at least this width are eligible
      max_width (int, optional): if present, only pixel maps of at most this width are eligible
      min_height (int, optional): if present, only pixel maps of at least this height are eligible
      max_height (int, optional): if present, only pixel maps of at most this height are eligible
      origin (float pair, optional): x, y of south west corner of area covered by pixel rectangle,
         in local crs of grid; if None then the minimum point of the grid's box is used and border
         applied
      dx (float, optional): the size (west to east) of a pixel, in locel crs of grid; if dx and dy
         are both None, they are computed from the grid's box, border, width and height
      dy (float, optional): the size (south to north) of a pixel, in locel crs; if only one of dx
         or dy is given, the other is set the same; if both None they are computed (see dx)
      border (float, default 0.0): a border width, in local crs of grid, applied around bounding box
         of grid when computing origin and/or dx & dy; ignored if those values are all supplied
      k0 (int, default 0): the layer to create a 2D pixel map for
      vertical_ref (string, default 'top'): 'top' or 'base'

   returns:
      resqpy.property.PropertyCollection holding at least one pixel map property

   notes:
      if exact_resolution is True, the individual min_ and max_ arguments are ignored;
      the origin, dx, dy, border, k0 & vertical_ref arguments are only used if a new pixel map is created;
      the epc is not rewritten by this function, even if new parts have been introduced and the hdf5 written to;
      calling code should use model.store_epc(only_if_modifed = True)
   """

   log.debug(f'finding or making pixel map with width, height: {width}, {height}')
   log.debug(f'width range: {min_width} to {max_width}; height range: {min_height} to {max_height}')

   if exact_resolution:
      min_width = max_width = width
      min_height = max_height = height

   pm_collection = find_pixel_maps(grid, min_width = min_width, max_width = max_width,
                                   min_height = min_height, max_height = max_height)

   if pm_collection is not None and pm_collection.number_of_parts() > 0: return pm_collection

   _, pm_collection = make_pixel_map(grid, width, height, origin = origin, dx = dx, dy = dy, border = border,
                                     k0 = k0, vertical_ref = vertical_ref, trim = trim)

   return pm_collection


def get_pixel_map(grid, width, height, exact_resolution = False,
                  min_width = None, max_width = None, min_height = None, max_height = None,
                  origin = None, dx = None, dy = None, border = 0.0,
                  k0 = None, vertical_ref = 'top', trim = True):
   """Finds or makes a suitable pixel map and returns numpy int array holding map.

   arguments:
      width (int): the width of the pixel rectangle (number of pixels)
      height (int): the height of the pixel rectangle (number of pixels)
      exact_resolution (bool, default False): if True, equivalent to setting the following min & max
         arguments equal to width and to height, forcing an exact match to the given resolution
      min_width (int, optional): if present, only pixel maps of at least this width are eligible
      max_width (int, optional): if present, only pixel maps of at most this width are eligible
      min_height (int, optional): if present, only pixel maps of at least this height are eligible
      max_height (int, optional): if present, only pixel maps of at most this height are eligible
      origin (float pair, optional): x, y of south west corner of area covered by pixel rectangle,
         in local crs of grid; if None then the minimum point of the grid's box is used and border
         applied
      dx (float, optional): the size (west to east) of a pixel, in locel crs of grid; if dx and dy
         are both None, they are computed from the grid's box, border, width and height
      dy (float, optional): the size (south to north) of a pixel, in locel crs; if only one of dx
         or dy is given, the other is set the same; if both None they are computed (see dx)
      border (float, default 0.0): a border width, in local crs of grid, applied around bounding box
         of grid when computing origin and/or dx & dy; ignored if those values are all supplied
      k0 (int, default None): the layer to create a 2D pixel map for
      vertical_ref (string, default 'top'): 'top' or 'base'
      trim (boolean, default True): if True, width or height may be trimmed to leave equal border

   returns:
      resqpy.property.PropertyCollection holding at least one pixel map property

   notes:
      if exact_resolution is True, the individual min_ and max_ arguments are ignored;
      the origin, dx, dy, border, k0 & vertical_ref arguments are only used if a new pixel map is created;
      the epc is not rewritten by this function, even if new parts have been introduced and the hdf5 written to;
      calling code should use model.store_epc(only_if_modifed = True);
      if trim is True, the resulting map may be smaller than min_height or min_width
   """

   log.debug(f'getting pixel map with width, height: {width}, {height}')

   pm_collection = find_or_make_pixel_map(grid, width, height, exact_resolution = exact_resolution,
                                          min_width = min_width, max_width = max_width,
                                          min_height = min_height, max_height = max_height,
                                          origin = origin, dx = dx, dy = dy, border = border,
                                          k0 = k0, vertical_ref = vertical_ref, trim = trim)

   pm_count = pm_collection.number_of_parts()
   assert pm_collection is not None and pm_count > 0
   pm_parts = pm_collection.parts()

   part = pm_parts[0]  # arbitrarily pick the first qualifying pixel map
   pixel_map = pm_collection.cached_part_array_ref(part, masked = False)
   assert pixel_map is not None

   return pixel_map


def width_and_height_of_pixel_map(pixel_map):
   """Returns the resolution of the pixel map as two integers being width, height."""

   assert pixel_map.ndim == 3 and pixel_map.shape[2] == 2
   height, width = pixel_map.shape[:2]
   return width, height


def colour_array(norm_prop, colour_map):
   """For a normalized float array, returns a numpy uint8 array with extent of extra last axis being 3 (RGB).

   arguments:
      norm_prop (numpy float array): array of data normalised to values between 0.0 and 1.0
      colour_map (matplotlib.cm colormap): the colour map to use to represent the normalised data

   returns:
      numpy uint8 array with same shape as norm_prop except with an extra axis appended with extent 3;
      the returned array holds RGB values for each of the input normalised elements
   """

   return (255.0 * (colour_map(norm_prop)[..., :3] + 0.00392)).astype(np.uint8)


def map_image(pixel_map, c_array, k0 = 0, image_file = None, background = 'lightgrey', file_only = False,
              title = None, title_location = (10, 10), title_colour = 'black'):
   """Produce a map image, written to file in PNG format and returned in a form to be assigned to an Image widget.

   arguments:
      pixel_map (numpy int array): pixel mapping array as returned by get_pixel_map()
      c_array (numpy uint8 array): RGB data for each cell in the source grid property, as returned by colour_array()
      k0 (int, default 0): the layer number to use from c_array, if for 3D property data
      image_file (string): the filename to write the PNG format image to; if None, a temporary file is written
         and removed immediately
      background (string, default 'lightgrey'): the matplotlib color to use for the background (unmapped pixels)
      file_only (boolean, default False): if True, the PNG file is written but None is returned
      title (string, optional): if present, text to be added to the map image
      title_location (pair of ints, default (10, 10)): pixel location to place origin of title, as (right, down)
         from top left corner of image; ignored if no title
      title_colour (string, default 'black'): the colour to stroke the title with; ignored if no title

   returns:
      image data suitable for assigning to the value attribute of an Image widget with a format of 'png',
      or None if file_only is True
   """

   assert pixel_map.ndim == 3 and pixel_map.shape[-1] == 2
   assert image_file or not file_only
   assert c_array.ndim == 4 or (c_array.ndim == 3 and k0 == 0)
   assert c_array.shape[-1] == 3   # todo: could allow for alpha channel?

   if c_array.ndim == 3: c_array = c_array.reshape(tuple([1] + list(c_array.shape)))
   assert 0 <= k0 < c_array.shape[0]

   remove_file = False
   if not image_file:
      image_file = '/tmp/map_' + str(bu.new_uuid()) + '.png'
      remove_file = True

   height, width = pixel_map.shape[:2]

   # create a new image using the Python Image Library package
   im = Image.new('RGB', (width, height), background)

   # get reference to actual pixels for image
   pixels = im.load()

   # each pixel has to be set individually!
   for row in range(height):
      pix_row = height - row - 1   # image starts at top left, whilst our pixel has origin bottom left
      for col in range(width):
          cell = pixel_map[row, col]
          if cell[0] < 0: continue  # pixel does not map to any IJ column, leave as background
          # colour-in pixel: RGB values must be presented as a tuple of 3 ints
          pixels[col, pix_row] = tuple(c_array[k0, cell[0], cell[1], :])

   if title:
      _add_text(im, title_location, title, title_colour)

   # save image to disc in PNG format
   im.save(image_file)

   png_image = None

   if not file_only:
      with open(image_file, "rb") as fh:
         png_image = fh.read()

   if remove_file: os.remove(image_file)

   return png_image


def _add_text(im, location, message, colour):
    """Add message text to image."""

    dim = ImageDraw.Draw(im)
    dim.text(location, message, fill = colour)




import pytest

import resqpy.organize as rqo
from resqpy.model import Model


# Test saving and loading from disk
@pytest.mark.parametrize("cls, data", [
    (
        rqo.OrganizationFeature,
        dict(feature_name='hello', organization_kind='stratigraphic'),
    ),
    (
        rqo.GeobodyFeature,
        dict(feature_name='hi'),
    ),
    (
        rqo.BoundaryFeature,
        dict(feature_name='foobar'),
    ),
    (
        rqo.FrontierFeature,
        dict(feature_name='foobar'),
    ),
    (
        rqo.GeologicUnitFeature,
        dict(feature_name='foobar'),
    ),
    (
        rqo.FluidBoundaryFeature,
        dict(feature_name='foobar', kind='gas oil contact'),
    ),
    (
        rqo.TectonicBoundaryFeature,
        dict(feature_name='foobar', kind='fracture'),
    ),
    (
        rqo.GeneticBoundaryFeature,
        dict(feature_name='foobar', kind='geobody boundary'),  # Age as well?
    ),
    (
        rqo.WellboreFeature,
        dict(feature_name='foobar'),
    ),
])
def test_organize_classes(tmp_model, cls, data):

    # Load example model from a fixture
    model = tmp_model
    epc = model.epc_file

    # Create the feature
    obj = cls(parent_model=model, **data)
    uuid = obj.uuid

    # Save to disk
    obj.create_xml()
    model.store_epc()
    model.h5_release()

    # Reload from disk
    del model, obj
    model2 = Model(epc_file=epc)
    obj2 = cls(parent_model=model2, uuid=uuid)

    # Check all attributes were loaded correctly
    for key, expected_value in data.items():
        assert getattr(obj2, key) == expected_value, f"Error for {key}"


def test_RockFluidUnitFeature(tmp_model):

    # Create the features
    top = rqo.BoundaryFeature(tmp_model, feature_name='the top')
    base = rqo.BoundaryFeature(tmp_model, feature_name='the base')
    rfuf_1 = rqo.RockFluidUnitFeature(
        parent_model=tmp_model, feature_name='foobar', phase='seal',
        top_boundary_feature=top, base_boundary_feature=base
    )
    uuid = rfuf_1.uuid

    # Save to disk
    top.create_xml()
    base.create_xml()
    rfuf_1.create_xml()
    tmp_model.store_epc()

    # Reload from disk
    rfuf_2 = rqo.RockFluidUnitFeature(parent_model=tmp_model, uuid=uuid)

    # Check properties the same
    assert rfuf_2.feature_name == 'foobar'
    assert rfuf_2.phase == 'seal'
    assert rfuf_2.top_boundary_feature.feature_name == 'the top'
    assert rfuf_2.base_boundary_feature.feature_name == 'the base'


def test_FaultInterp(tmp_model):
    
    title = "not my fault"
    tect_boundary = rqo.TectonicBoundaryFeature(tmp_model, kind='fault')
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        tectonic_boundary_feature=tect_boundary,
        title=title,
        domain="depth",
        is_normal=True,
        maximum_throw=3,
    )

    tect_boundary.create_xml()
    fault_interp.create_xml()

    fault_interp_2 = rqo.FaultInterpretation(tmp_model, uuid=fault_interp.uuid)
    assert fault_interp_2.title == title
    assert fault_interp_2.maximum_throw == 3

def test_EarthModel(tmp_model):

    title='gaia'
    org_feat = rqo.OrganizationFeature(tmp_model, feature_name='marie kondo', organization_kind="earth model")
    em1 = rqo.EarthModelInterpretation(tmp_model, title=title, organization_feature=org_feat)
    
    org_feat.create_xml()
    em1.create_xml()

    em2 = rqo.EarthModelInterpretation(tmp_model, uuid=em1.uuid)
    assert em2.title == title


def test_Horizon(tmp_model):

    gen = rqo.GeneticBoundaryFeature(tmp_model, kind='horizon')
    hor = rqo.HorizonInterpretation(
        tmp_model, genetic_boundary_feature=gen, sequence_stratigraphy_surface='maximum flooding'
    )

    gen.create_xml()
    hor.create_xml()

    hor2 = rqo.HorizonInterpretation(tmp_model, uuid=hor.uuid)
    assert hor2.sequence_stratigraphy_surface == hor.sequence_stratigraphy_surface

def test_GeobodyBoundary(tmp_model):

    gen = rqo.GeneticBoundaryFeature(tmp_model, kind='geobody boundary')
    gb = rqo.GeobodyBoundaryInterpretation(
        tmp_model, genetic_boundary_feature=gen
    )

    gen.create_xml()
    gb.create_xml()
    gb2 = rqo.GeobodyBoundaryInterpretation(tmp_model, uuid=gb.uuid)
    assert gb == gb2



"""vdb.py: Module providing functions for reading from VDB datasets."""

version = '15th March 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging
log = logging.getLogger(__name__)
log.debug('vdb.py version ' + version)

import os
import glob
import numpy as np
import zipfile as zf
from struct import unpack

import resqpy.olio.xml_et as rqet
import resqpy.olio.grid_functions as gf


null_uint32 = 4294967295        # -1 if interpreted as int32

key_dict = { # vdb key character mapping to: (numpy_dtype, size_in_bytes, unpack_format_ch)
   'R': ('float32', 4, 'f'),
   'D': ('float64', 8, 'd'),
#   'D': ('int64',   8, 'i'),
   'I': ('int32',   4, 'i'),
   'C': (None,      1, 'c'),    # could map to numpy 'byte' but seems to be used for strings
   'P': ('uint32',  4, 'I'),
   'K': (None,      8, 'c'),    # don't store in numpy format; 8 character strings
   'X': (None,      0, 'c')}    # used for invalid code character (non-ascii)

init_not_packed = ['DAD', 'KID', 'UID', 'UNPACK']

def coerce(a, dtype):
   """Returns a version of numpy array a with elements coerced to dtype.

      :meta private:
   """
   if dtype is None or a.dtype == dtype: return a
   b = np.empty(a.shape, dtype = dtype)
   b[:] = a
   return b


def ensemble_vdb_list(run_dir, sort_list = True):
   """Returns a sorted list of vdb paths found in the directory tree under run_dir."""

   ensemble_list = []

   def recursive_vdb_list(dir):
      nonlocal ensemble_list
      for entry in os.scandir(dir):
         if not entry.is_dir(): continue
         if entry.name.endswith('.vdb') or entry.name.endswith('.vdb.zip'):
            ensemble_list.append(entry.path)
            continue
         elif entry.name.endswith('.rst'):
            continue # optimisation
         recursive_vdb_list(entry.path)

   def cmp_to_key(mycmp):
      'Convert a cmp= function into a key= function'
      class K:

         def __init__(self, obj, *args):
            self.obj = obj

         def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0

         def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0

         def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0

         def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0

         def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0

         def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0

      return K

   def comparison(a, b):
      pa = 0
      pb = 0
      while True:
         sa = pa
         sb = pb
         if pa >= len(a) and pb >= len(b): return 0
         while pa < len(a) and not a[pa].isdigit(): pa += 1
         while pb < len(b) and not b[pb].isdigit(): pb += 1
         if a[sa:pa].lower() < b[sb:pb].lower(): return -1
         if a[sa:pa].lower() > b[sb:pb].lower(): return 1
         if pa >= len(a) and pb >= len(b): return 0
         if pa >= len(a): return -1
         if pb >= len(b): return 1
         sa = pa
         sb = pb
         while pa < len(a) and a[pa].isdigit(): pa += 1
         while pb < len(b) and b[pb].isdigit(): pb += 1
         ia = int(a[sa:pa])
         ib = int(b[sb:pb])
         if ia < ib: return -1
         if ia > ib: return 1
      a_low = a.lower()
      b_low = b.lower()
      if a_low < b_low: return -1
      if a_low > b_low: return 1
      return 0

   recursive_vdb_list(run_dir)
   if sort_list:
      sorted_list = sorted(ensemble_list, key = cmp_to_key(comparison))
      ensemble_list = sorted_list
   return ensemble_list


class Header():
   """Internal class for handling a Header record in a vdb file."""

   def __init__(self, fp, place):
      """Creates a new Header record object."""

      fp.seek(place)
      block = fp.read(22)
#      log.debug(f'Header block at place {place} returned {len(block)} bytes')
      self.previous, self.next, c, self.bytes_per_item, self.number_of_items, self.first_fragment, self.max_items =  \
         unpack('=IIcBIII', block)
      try:
         self.item_type = c.decode()
      except:
         self.item_type = 'X'      # non-ascii character!
      self.data_place = place + 22
#      log.debug('   header previous: ' + str(self.previous))
#      log.debug('   header next: ' + str(self.next))
#      log.debug('   header item type: ' + str(self.item_type))
#      log.debug('   header bytes per item: ' + str(self.bytes_per_item))
#      log.debug('   header number of items: ' + str(self.number_of_items))
#      log.debug('   header first fragment: ' + str(self.first_fragment))
#      log.debug('   header max items: ' + str(self.max_items))
#      log.debug('   header data place: ' + str(self.data_place))


class FragmentHeader():
   """Internal class for handling a Fragment Header record in a vdb file."""

   def __init__(self, fp, place):
      """Creates a new Fragment Header record object."""

#      log.debug(f'FragmentHeader init at place {place}')
      fp.seek(place)
      block = fp.read(8)
      self.next, self.number_of_items = unpack('=II', block)
      self.data_place = place + 8
#      log.debug(f'   fragment header number of items {self.number_of_items}; data place {self.data_place}; next {self.next}')


class RawData():
   """Internal class for handling Raw Data records in a vdb file."""

   def __init__(self, fp, place, item_type, count, max_count):
      """Creates a new Raw Data record object."""

      if max_count is not None and count > max_count: count = max_count
      self.a = None
      self.c = None
      dtype, byte_size, form_ch = key_dict[item_type]
#      log.debug('raw data call: place {}; item type {}; count {}; dtype {}, byte size {}, form ch {}'.format(
#                place, item_type, count, dtype, byte_size, form_ch))
      fp.seek(place)
      if dtype is None:
         if form_ch == 'c':
            chars = fp.read(count * byte_size)
            if item_type == 'C':
               self.c = chars.decode()
            else:
               self.c = []
               for i in range(count):
                  self.c.append(chars[i * byte_size : (i + 1) * byte_size].decode().strip().upper())
         else: # shouldn't come into play
            block = fp.read(count * byte_size)
            form = '=' + str(count) + form_ch
            self.c = unpack(form, block)
      elif dtype == 'float64':                        # try tentative 32bit word swap
         b = fp.read(count * 8)
#         c = b''
#         for d in range(count):
#            c += b[4*d+4:4*d+8] + b[4*d:4*d+4]
         nda = np.ndarray((count, 2), dtype = 'int32', buffer = b)
         self.a = np.empty((count, 2), dtype = int)
         self.a[:, :] = nda[:, :]
      else:
         self.a = np.empty((count,), dtype = dtype)
         self.a.data = fp.read(count * byte_size)     # todo: check C ordering; endianess etc.


class Data():
   """Internal class for handling Data records in a vdb file."""

   def __init__(self, fp, header):
      """Creates a new Data object."""

      if header is None:
         self.a = None
         self.c = None
      else:
#         log.debug(f'Data init at place {header.data_place}; type {header.item_type}; number of items {header.number_of_items}')
         raw = RawData(fp, header.data_place, header.item_type, header.number_of_items, header.max_items)
         if raw is not None and raw.a is not None and raw.a.size == 1 and raw.a.dtype == 'int32' and header.next != null_uint32:
#            log.debug('   skipping integer value of ' + str(raw.a[0]))
            next_head = Header(fp, header.next)
            raw = RawData(fp, next_head.data_place, next_head.item_type, next_head.number_of_items, next_head.max_items)
            header = next_head  # making this up as I go along
         self.a = raw.a
         self.c = raw.c
         if header.first_fragment != null_uint32:
#            log.debug(f'   chaining from {header.first_fragment}')
            chain = FragmentChain(fp, header.first_fragment, header)
            if self.a is not None: self.a = np.append(self.a, chain.a)
            if self.c is not None: self.c += chain.c
   #   if self.c is not None:
   #      log.debug(f'   data c {self.c}')
   #   elif self.a is not None:
   #      log.debug(f'   data a {self.a}')


class Fragment():
   """Internal class for handling an individual Fragment record in a vdb file."""

   def __init__(self, fp, place, header):
      """Creates a new Fragment record object."""

#      log.debug(f'Fragment init at place {place}')
      self.head = FragmentHeader(fp, place)
      if self.head.number_of_items == 0:
#         log.debug('   zero items in fragment')
         self.c = None
         self.a = None
      else:
         raw = RawData(fp, self.head.data_place, header.item_type, self.head.number_of_items, header.max_items)
         self.a = raw.a
         self.c = raw.c
#      if self.c is not None:
#         log.debug(f'   fragment c [{self.c}]')


class FragmentChain():
   """Internal class for handling a chain of Fragment records in a vdb file."""

   def __init__(self, fp, place, header):  # returns either 1D numpy array (numeric data) or list of 8
      """Creates a new Fragment Chain object."""

      self.a = None
      self.c = None
#      log.debug(f'FragmentChain init at place {place}')
      while place != null_uint32:
         fragment = Fragment(fp, place, header)
         if fragment.head.number_of_items > 0:
            if fragment.a is not None:
               if self.a is None: self.a = fragment.a
               else: self.a = np.concatenate(self.a, fragment.a)
            if fragment.c:
               if self.c is None: self.c = []
               if isinstance(fragment.c, str):
                  self.c.append(fragment.c)
               elif isinstance(fragment.c, list):
                  for item in fragment.c:
                     if item.isascii(): self.c.append(item)
         place = fragment.head.next
      assert self.c is None or self.a is None, 'mixture of character and numeric data in fragment chain'
   # todo: check number of elements matches header info?


class KP():
   """Internal class for a (Key, Pointer) record in a vdb file."""

   def __init__(self, fp, place = 4):
      """Creates a new (Key, Pointer) record object."""

#      log.debug(f'KP init at place {place}')
      self.k_head = Header(fp, place)
      assert self.k_head.item_type == 'K', 'did not find expected Key header'
      assert self.k_head.bytes_per_item == 8, 'bytes per item not 8 in Key header'
      assert self.k_head.number_of_items > 0, 'zero items in Key header'
      place = self.k_head.next
      assert place != null_uint32, 'no next header in Key header'
      self.p_head = Header(fp, place)
      assert self.p_head.item_type == 'P', 'did not find expected Pointer header'
      assert self.p_head.bytes_per_item == 4, 'bytes per item not 4 in Pointer header'
      assert self.p_head.number_of_items == self.k_head.number_of_items,  \
                'number of items in Pointer header does not match number in Key header'
      self.keywords = Data(fp, self.k_head).c
      assert self.keywords is not None and isinstance(self.keywords, list), 'list of keywords not extracted from Key record'
      self.pointers = Data(fp, self.p_head).a
      assert self.pointers is not None and self.pointers.dtype == 'uint32'
      self.fp = fp

   def header_place_for_key(self, key, search = False):
      """Returns file position pointer for a given key."""

      key = key.strip().upper()
      if key in self.keywords:
         return self.pointers[self.keywords.index(key)]
      if not search: return None
      for head_key in self.keywords:
         sub_head_place = self.pointers[self.keywords.index(head_key)]
         try:
            sub_kp = KP(self.fp, sub_head_place)
            sub_place = sub_kp.header_place_for_key(key, search = True)
            if sub_place is not None: return sub_place
         except AssertionError:
            pass  # probably not a K type subsidiary
         except:
            raise
      return None

   def head_for_key(self, key, search = False):
      """Returns a Header record object for the given key."""

      head_place = self.header_place_for_key(key, search = search)
      if head_place is None: return None
      head = Header(self.fp, head_place)
      if head.item_type == 'X': return None    # invalid (non-ascii) item type character
      return head

   def data_for_key(self, key, search = False):
      """Returns a Data object for the given key."""

      head = self.head_for_key(key, search = search)
      if head is None: return None
      return Data(self.fp, head)

   def key_list(self, filter = False):
      """Returns a list of keys."""

#      return Data(self.fp, self.k_head).c
      if not filter: return self.keywords
      filtered_list = []
      for key in self.keywords:
#        log.debug(f'key_list raw entry: {key}')
        if key.isalnum(): filtered_list.append(key)
      return filtered_list

   def sub_key_list(self, keyword, filter = False):
      """Returns a list of keys subordinate to keyword."""

      assert keyword in self.key_list(), 'keyword not present: ' + keyword
      sub_head_place = self.pointers[self.keywords.index(keyword)]
      sub_kp = KP(self.fp, sub_head_place)
      return sub_kp.key_list(filter = filter)


class VDB():
   """Class for handling a vdb, particularly to support import of grid and properties."""

   def __init__(self, path):
      """Initialises a VDB object and associates it with the given vdb directory path."""

      self.path = None       # internal zip path of vdb in the case of a zip file
      self.zip_file = None
      self.zipped = zf.is_zipfile(path)
      if self.zipped:
         self.zip_file = path
         with zf.ZipFile(path) as zfp:
            zip_list = zfp.namelist()
            warn = False
            for name in zip_list:
               if name.endswith('.vdb/'):
                  if self.path is not None: warn = True
                  self.path = name[:-1]
               elif self.path is None and name == 'main.xml':
                  self.path = ''
         if warn: log.warning('more than one vdb in zip file: using last')
         if self.path is None:
            raise Exception('no vdb found in zip file (at top level)')
      else:
         self.path = path
         if not os.path.isdir(path):
            log.warning('vdb path is neither a directory nor a zip file: ' + str(path))
      self.case_list = None
      self.use_case = None
      self.grid_extents_kji = {}  # maps case name, grid_name to grid extent
      self.grid_packings = {}     # maps case name, grid_name to grid unpack array
      self.grid_lists = {}        # maps case name to grid name list
      self.cases()                # sets use_case to first case in list for vdb


   def cases(self):
      """Returns a list of simulation case strings as found in the main xml file."""

      if self.case_list is not None: return self.case_list
      try:
         xml_file = os.path.join(self.path, 'main.xml')
         if self.zipped:
            with zf.ZipFile(self.zip_file) as zfp:
               with zfp.open(xml_file) as fp:
                  tree = rqet.parse(fp)
                  root = tree.getroot()
         else:
            assert os.path.exists(xml_file), 'could not find vdb main xml file: ' + xml_file
            with open(xml_file, 'r') as fp:
               tree = rqet.parse(fp)
               root = tree.getroot()
         caselist = rqet.list_of_tag(rqet.find_tag(root, 'CASELIST'), 'CASE')
         self.case_list = []
         for case in caselist:
             self.case_list.append(str(case.attrib['Name']).strip())
         if self.use_case is None and len(self.case_list): self.use_case = self.case_list[0]
      except:
         log.exception('failed to extract case list')
      return self.case_list


   def set_use_case(self, case):
      """Sets the simulation case to use in other functions."""

      try:
         assert case in self.cases(), 'failed to find case ' + case + ' in list of vdb cases'
         self.use_case = case
      except:
         log.exception('failed to set use case')


   def print_header_tree(self, relative_path):
      """Low level: prints out the raw header tree found in the vdb file (for debugging)."""

      visited = []

      def print_tree(fp, place = 4, level = 0, dots = False, file_size = None):
         try:
            if file_size is None:
               fp.seek(0, 2)
               file_size = int(fp.tell())
            assert 0 <= int(place) < file_size, f'place {place} not within bounds of file of size {file_size}'
            assert place not in visited, 'circular pointers in vdb file header structure'
            visited.append(place)
            head = Header(fp, place = place)
            if head is None: raise ValueError()
            if dots: indent = ((level - 1) * 3) * ' ' + ' ..'
            else: indent = (level * 3) * ' '
            print(indent, head.item_type, ':', head.number_of_items, 'items at', head.bytes_per_item, 'bytes per item', end = '')
            if head.item_type == 'C':
               s = RawData(fp, head.data_place, 'C', head.number_of_items, head.max_items)
               if s is None or s.c is None: print(': ?', end = '')
               else: print(': "' + s.c + '"', end = '')
            if head.first_fragment != null_uint32: print(' with fragment(s)')
            else: print('')
            if head.item_type == 'P':
               pointers = RawData(fp, head.data_place, 'P', head.number_of_items, head.max_items)
               assert(pointers is not None and pointers.a is not None)
               for p in pointers.a:
                  if p in visited:
                     log.warning('breaking at circular pointer to place ' + str(p))
                     break
                  if int(p) == 0:
                     log.warning('pointer to place zero encountered')
                  elif int(p) > file_size:
                     log.warning('skipping pointer to place beyond end of file ' + str(p))
                  else:
                     print_tree(fp, place = p, level = level + 1, dots = True, file_size = file_size)
            if head.next != null_uint32:
               if not 0 <= int(head.next) < file_size:
                  log.warning(f'skipping next {head.next} outwith bounds of file of size {file_size}')
               else:
                  print_tree(fp, place = head.next, level = level, file_size = file_size)
         except:
            print('?')
            log.exception('failed in print tree for headers')
            raise

      try:
         path = os.path.join(self.path, self.use_case, relative_path)
         assert os.path.exists(path), 'failed to find vdb file ' + path
         print('Header tree for: ' + path)
         with open(path, 'rb') as fp:
            print_tree(fp)
      except:
         log.exception('failed to print header tree for relative file ' + relative_path)


   def print_key_tree(self, relative_path):
      """Low level: prints out the keyword tree found in the vdb file."""

      visited = []

      def print_tree(fp, place = 4, level = 0):
         try:
            if int(place) == 0:
               log.warning('returning due to place zero in vdb file key structure')
               return
            if place in visited:
               log.warning(f'returning at circular pointer at place {place} in vdb file key structure')
               return
            visited.append(place)
            kp = KP(fp, place = place)
         except AssertionError:
            return
         except Exception as e:
            log.error(str(e) + ' at place ' + str(place))
            raise
         key_list = kp.key_list()
         for key in key_list:
            print((level * 3) * ' ', key)
            print_tree(fp, kp.header_place_for_key(key, search = False), level + 1)

      try:
         path = os.path.join(self.path, self.use_case, relative_path)
         if self.zipped:
            with zf.ZipFile(self.zip_file) as zfp:
               with zfp.open(path) as fp:
                  print_tree(fp)
         else:
            assert os.path.exists(path), 'failed to find vdb file ' + str(path)
            with open(path, 'rb') as fp:
               print_tree(fp)
      except:
         log.exception('failed to print keyword tree for relative file ' + relative_path)


   def data_for_keyword(self, relative_path, keyword, search = True):
      """Reads data associated with a keyword from a vdb binary file; returns a numpy array (or string)."""

      try:
         path = os.path.join(self.path, self.use_case, relative_path)
         if self.zipped:
            with zf.ZipFile(self.zip_file) as zfp:
               with zfp.open(path) as fp:
                  assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
                  kp = KP(fp)
                  return kp.data_for_key(keyword, search = search)
         else:
            assert os.path.exists(path), 'failed to find vdb file ' + str(path)
            with open(path, 'rb') as fp:
               assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
               kp = KP(fp)
               return kp.data_for_key(keyword, search = search)
      except:
         log.exception('failed to read keyword data from vdb binary file: ' + path)
      return None


   def data_for_keyword_chain(self, relative_path, keyword_chain):
      """Follows a list of keywords down through hierarchy and returns the data as a numpy array (or string)."""

      def dfkc_fp(fp, path, keyword_chain):
         assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
         chain_list = list(keyword_chain)
         assert len(chain_list) > 0, 'empty chained keyword list'
         head_place = 4
         while len(chain_list) > 1:
            kp = KP(fp, place = head_place)
            head_place = kp.header_place_for_key(chain_list[0], search = False)
            assert head_place is not None, 'failed to find chained keyword: ' + chain_list[0]
            chain_list.pop(0)
         kp = KP(fp, place = head_place)
         return kp.data_for_key(chain_list[0], search = False)

      path = os.path.join(self.path, self.use_case, relative_path)
      try:
         if isinstance(keyword_chain, str): return self.data_for_keyword(relative_path, keyword_chain)
         if self.zipped:
            with zf.ZipFile(self.zip_file) as zfp:
               with zfp.open(path) as fp:
                  return dfkc_fp(fp, path, keyword_chain)
         else:
            assert os.path.exists(path), 'failed to find vdb file ' + str(path)
            with open(path, 'rb') as fp:
               return dfkc_fp(fp, path, keyword_chain)
      except:
         log.exception('failed to read chained keyword data from vdb binary file: ' + path)
      return None


   def set_extent_kji(self, extent_kji, use_case = None, grid_name = 'ROOT'):
      """Sets extent for one use case (defaults to current use case) as alternative to processing corp data."""

      assert len(extent_kji) == 3, 'triple integer required for extent_kji'
      if use_case is None: use_case = self.use_case
      assert use_case is not None, 'no use case for extent setting'
      self.grid_extents_kji[use_case, grid_name.upper()] = extent_kji


   def fetch_corp_patch(self, relative_path):
      """Loads one patch of grid corp data from one file in the vdb; returns (number of cells, 1D array)."""

      a = self.data_for_keyword(relative_path, 'CORP').a
      assert a is not None, 'failed to extract corp data from vdb relative path ' + relative_path
      element_count = a.size
      assert element_count > 0, 'corp data is empty (no numbers)'
      cells, remainder = divmod(element_count, 24)
      assert remainder == 0, 'number of numbers in corp data is not a multiple of 24'
      if a.dtype == 'float32':
         d = np.empty((element_count, ), dtype = 'float')   # double precision
         d[:] = a
         del a
         a = d
      return cells, a


   def list_of_grids(self):
      """Returns a list of grid names for which corp data exists in the vdb for the current use case."""

      if self.use_case is None: self.cases()
      if self.use_case in self.grid_lists: return self.grid_lists[self.use_case]
      glob_path = os.path.join(self.path, self.use_case, 'INIT', '*_corp.bin')
      if self.zipped:
         corp_list = self.zip_glob(glob_path)
      else:
         corp_list = glob.glob(glob_path)
      grid_name_list = []
      for corp_path in corp_list:
         _, corp_file_name = os.path.split(corp_path)
         grid_name = corp_file_name[:-9].upper()
         if grid_name == 'ROOT': grid_name_list.insert(0, 'ROOT')
         else: grid_name_list.append(grid_name)
      self.grid_lists[self.use_case] = grid_name_list
      return grid_name_list


   def root_corp(self):
      """Loads root grid corp data from vdb; returns pagoda style resequenced 7D numpy array of doubles."""

      return self.grid_corp('ROOT')


   def grid_corp(self, grid_name):
      """Loads corp data for named grid from vdb; returns pagoda style resequenced 7D numpy array of doubles."""
      try:
         if self.use_case is None: self.cases()
         grid_name = grid_name.upper()
         assert self.use_case is not None, 'no case found in vdb'
         relative_path = os.path.join('INIT', str(grid_name) + '_corp.bin')
         cells, a = self.fetch_corp_patch(relative_path)
         # check for extra corp patch files (used for big grids)
         glob_path = os.path.join(self.path, self.use_case, 'INIT', str(grid_name) + '_corp')
         glob_path += '_*.bin'
         if self.zipped:
            corp_patch_list = self.zip_glob(glob_path)
         else:
            corp_patch_list = glob.glob(glob_path)
         if len(corp_patch_list) > 0:
            sort_list = []
            for patch_name in corp_patch_list:
               layer_number = int(patch_name[patch_name.rfind('_') + 1 : -4])
               sort_list.append((layer_number, patch_name))
            sort_list.sort()
            for _, patch_name in sort_list:
               patch_relative_path = patch_name[patch_name.rfind('INIT'):]
               patch_cells, patch = self.fetch_corp_patch(patch_relative_path)
               a = np.append(a, patch)
               cells += patch_cells
         ap = a.reshape((1, 1, cells, 2, 2, 2, 3))
         gf.resequence_nexus_corp(ap)                  # move from Nexus corp ordering to Pagoda ordering
         if (self.use_case, grid_name) in self.grid_extents_kji:
            extent_kji = self.grid_extents_kji[self.use_case, grid_name]
            assert cells == extent_kji[0] * extent_kji[1] * extent_kji[2], 'corp extent mismatch in vdb'
         else:
            extent_kji = gf.determine_corp_extent(ap)
            if extent_kji is None:
               log.warning('failed to determine extent of root grid from corp data')
               extent_kji = (1, 1, cells)
            else:
               self.grid_extents_kji[self.use_case, grid_name] = extent_kji
         shape_7d = (extent_kji[0], extent_kji[1], extent_kji[2], 2, 2, 2, 3)
         return ap.reshape(shape_7d)
      except:
         log.exception('failed to extract root grid corp data from vdb')
      return None


   def load_init_mapdata_array(self, file, keyword, dtype = None, unpack = False, grid_name = 'ROOT'):
      """Loads an INIT MAPDATA array from vdb; returns 3D numpy array coerced to dtype (if not None)."""

      try:
         if self.use_case is None: self.cases()
         assert self.use_case is not None, 'no case found in vdb'
         relative_path = os.path.join('INIT', 'MAPDATA', file)
         a = self.data_for_keyword(relative_path, keyword, search = True).a
         assert a is not None, 'failed to extract data for keyword ' + keyword + ' from vdb relative path ' + relative_path
         if unpack:
            un = self.grid_unpack(grid_name)
            if un is not None:
               null_start = np.zeros((a.size + 1, ), dtype = a.dtype)  # better to use Nan for null value?
               null_start[1:] = a
               a = null_start[un]
         a = self.grid_shaped(grid_name, coerce(a, dtype))
         return a
      except:
         log.exception('failed to extract data from vdb for grid ' + str(grid_name))
      return None


   def load_recurrent_mapdata_array(self, file, keyword, dtype = None, unpack = False, grid_name = 'ROOT'):
      """Loads a RECUR MAPDATA array from vdb; returns 3D numpy array coerced to dtype (if not None)."""

      try:
         if not keyword or not keyword.isalnum():
            log.warning('ignoring attempt to load recurrent vdb mapdata for corrupt keyword')
            return None
         if self.use_case is None: self.cases()
         assert self.use_case is not None, 'no case found in vdb'
         relative_path = os.path.join('RECUR', 'MAPDATA', file)
         data = self.data_for_keyword(relative_path, keyword, search = True)
         if not data: return None
         a = data.a
         if a is None: return None
         if unpack:
            un = self.grid_unpack(grid_name)
            if un is not None:
               null_start = np.zeros((a.size + 1, ), dtype = a.dtype)  # better to use Nan for null value?
               null_start[1:] = a
               a = null_start[un]
         a = self.grid_shaped(grid_name, coerce(a, dtype))
         return a
      except:
         log.exception('failed to extract grid recurrent data from vdb for keyword: ' + keyword)
      return None


   def root_dad(self):
      """Loads and returns the IROOTDAD array from vdb; returns 3D numpy int32 array."""

      return self.grid_dad('ROOT')


   def grid_dad(self, grid_name):
      """Loads and returns the DAD array from vdb for the named grid; returns 3D numpy int32 array."""

      # todo: check file naming and values for LGRs
      grid_name = grid_name.upper
      # DAD data appear to be unique cell ids, in usual sequence, starting at 1, same as UID
      return self.load_init_mapdata_array(file = 'I' + grid_name + 'DAD.bin', keyword = 'DAD',
                                          dtype = 'int32', unpack = False, grid_name = grid_name)


   def root_kid(self):
      """Loads and returns the IROOTKID array from vdb; returns 3D numpy int32 array (can be inactive cell mask)."""

      return self.grid_kid('ROOT')


   def grid_kid(self, grid_name):
      """Loads and returns the IROOTKID array from vdb; returns 3D numpy int32 array (can be inactive cell mask)."""

      grid_name = grid_name.upper()
      # see comments in grid_kid_inactive_mask() for KID values
      return self.load_init_mapdata_array(file = 'I' + grid_name + 'KID.bin', keyword = 'KID',
                                          dtype = 'int32', unpack = False, grid_name = grid_name)


   def root_kid_inactive_mask(self):
      """Loads the IROOTKID array and returns boolean mask of cells inactive in ROOT grid."""

      return self.grid_kid_inactive_mask('ROOT')


   def grid_kid_inactive_mask(self, grid_name):
      """Loads the KID array for the named grid and returns boolean mask of cells inactive in grid."""

      # KID values:
      # 0 - active in this grid
      # >0 - (in ROOT grid) cell inactive in ROOT as assigned to an LGR (KID value is LGR number)
      # -1 – inactive for normal reasons (zero pore volume or DEADCELL set)
      # -2 - (in LGR grids) cell omitted from LGR (possibly omitted as inactive in parent grid?)
      # -3 - inactive due to coarsening (coarsening repurposes one cell in parent grid for each coarsened cell)
      i = self.grid_kid(grid_name)
      if i is None: return None
      b = np.empty(i.shape, dtype = 'bool')
      b[:] = (i != 0)
      return b


   def root_uid(self):
      """Loads and returns the IROOTUID array from vdb; returns 3D numpy int32 array."""

      return self.grid_uid('ROOT')


   def grid_uid(self, grid_name):
      """Loads and returns the UID array from vdb for the named grid; returns 3D numpy int32 array."""

      grid_name = grid_name.upper()
      # UID data appear to be unique cell ids, in usual sequence, starting at 1, same as DAD
      return self.load_init_mapdata_array(file = 'I' + grid_name + 'UID.bin', keyword = 'UID',
                                          dtype = 'int32', unpack = False, grid_name = grid_name)


   def root_unpack(self):
      """Loads and returns the IROOTUNPACK array from vdb; returns 3D numpy int32 array."""

      return self.grid_unpack('ROOT')


   def grid_unpack(self, grid_name):
      """Loads and returns the IROOTUNPACK array from vdb; returns 3D numpy int32 array."""

      grid_name = grid_name.upper()
      if self.use_case is not None and (self.use_case, grid_name) in self.grid_packings.keys():
         return self.grid_packings[self.use_case, grid_name]
      # UNPACK data appear to be index into packed array for each cell, zero for inactive/absent
      un = self.load_init_mapdata_array(file = 'I' + grid_name + 'UNPACK.bin', keyword = 'UNPACK',
                                        dtype = 'int32', unpack = False, grid_name = grid_name)
      if self.use_case is not None:
         self.grid_packings[self.use_case, grid_name] = un   # cache unpack array for later packing & unpacking operations
      return un


   def list_of_static_properties(self):
      """Returns list of static property keywords present in the vdb for ROOT."""

      return self.grid_list_of_static_properties('ROOT')


   def grid_list_of_static_properties(self, grid_name):
      """Returns list of static property keywords present in the vdb for named grid."""

      grid_name = grid_name.upper()
      glob_path = os.path.join(self.path, self.use_case, 'INIT', 'MAPDATA', 'I' + grid_name + '*.bin')
      if self.zipped:
         file_list = self.zip_glob(glob_path)
      else:
         file_list = glob.glob(glob_path)
      keyword_list = []
      for p in file_list:
         _, f = os.path.split(p)
         keyword = f[1 + len(grid_name):-4]
         keyword_list.append(keyword)
      return keyword_list


   def root_static_property(self, keyword, dtype = None, unpack = None):
      """Loads and returns a ROOT static property array."""

      return self.grid_static_property('ROOT'. keyword, dtype = dtype, unpack = unpack)


   def grid_static_property(self, grid_name, keyword, dtype = None, unpack = None):
      """Loads and returns a static property array for named grid."""

      grid_name = grid_name.upper()
      keyword = keyword.strip().upper()
      if unpack is None: unpack = (keyword not in init_not_packed)
      return self.load_init_mapdata_array(file = 'I' + grid_name + keyword + '.bin', keyword = keyword,
                                          dtype = dtype, unpack = unpack, grid_name = grid_name)


   def list_of_timesteps(self):
      """Returns a list of integer timesteps for which a ROOT recurrent mapdata file exists."""

      return self.grid_list_of_timesteps('ROOT')


   def grid_list_of_timesteps(self, grid_name):
      """Returns a list of integer timesteps for which a recurrent mapdata file for the named grid exists."""

      # todo: check file naming for LGR recurrent properties
      grid_name = grid_name.upper()
      glob_path = os.path.join(self.path, self.use_case, 'RECUR', 'MAPDATA', 'R' + grid_name + '_*.bin')
      if self.zipped:
         file_list = self.zip_glob(glob_path)
      else:
         file_list = glob.glob(glob_path)
      timestep_list = []
      for p in file_list:
         timestep_list.append(int(p[p.rfind('_')+1:-4]))
      return sorted(timestep_list)


   def list_of_recurrent_properties(self, timestep):
      """Returns list of recurrent property keywords present in the vdb for given timestep."""

      return self.grid_list_of_recurrent_properties('ROOT', timestep)


   def grid_list_of_recurrent_properties(self, grid_name, timestep):
      """Returns list of recurrent property keywords present in the vdb for named grid for given timestep."""

      def keyword_list_from_fp(fp, path):
         assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
         kp = KP(fp)
         keyword_list = kp.sub_key_list('MAPDATA', filter = True)
         try:
            keyword_list.remove('LASTMOD')
         except:
            pass
         bad_key_indices = []
         for i in range(len(keyword_list)):
            key = keyword_list[i]
            if not key or not key.isalnum(): bad_key_indices.append(i)
         if len(bad_key_indices):
            log.warning(str(len(bad_key_indices)) + ' non-ascii keywords ignored in recurrent file ' + path)
            bad_key_indices.reverse()
            for i in bad_key_indices: keyword_list.pop(i)
         return keyword_list

      # todo: check file naming for LGR recurrent properties
      grid_name = grid_name.upper()
      try:
         path = os.path.join(self.path, self.use_case, 'RECUR', 'MAPDATA', 'R' + grid_name + '_' + str(timestep) + '.bin')
         if self.zipped:
            with zf.ZipFile(self.zip_file) as zfp:
               with zfp.open(path) as fp:
                  return keyword_list_from_fp(fp, path)
         else:
            assert os.path.exists(path), 'failed to find vdb file ' + str(path)
            with open(path, 'rb') as fp:
               return keyword_list_from_fp(fp, path)
      except:
         log.exception('failed to read keyword data from vdb binary file: ' + path)
      return None


   def root_recurrent_property_for_timestep(self, keyword, timestep, dtype = None, unpack = True):
      """Loads and returns a ROOT recurrent property array for one timestep."""

      return self.grid_recurrent_property_for_timestep('ROOT', keyword, timestep, dtype = dtype, unpack = unpack)


   def grid_recurrent_property_for_timestep(self, grid_name, keyword, timestep, dtype = None, unpack = True):
      """Loads and returns a recurrent property array for named grid for one timestep."""

      # todo: check file naming convention for LGRs
      grid_name = grid_name.upper()
      keyword = keyword.strip().upper()
      return self.load_recurrent_mapdata_array(file = 'R' + grid_name + '_' + str(timestep) + '.bin',
                                               keyword = keyword, dtype = dtype, unpack = unpack, grid_name = grid_name)


   def header_place_for_keyword(self, relative_path, keyword, search = True):
      """Low level function to return file position for header relating to given keyword."""

      try:
         path = os.path.join(self.path, relative_path)
         assert os.path.exists(path), 'failed to find vdb file ' + str(path)
         with open(path, 'rb') as fp:
            assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
            kp = KP(fp)
            return kp.header_place_for_key(keyword, search = search)
      except:
         log.exception('failed to read keyword data from vdb binary file: ' + path)
      return None


   def root_shaped(self, a):
      """Returns array reshaped to root grid extent for current use case, if known; otherwise unchanged."""

      return self.grid_shaped('ROOT', a)


   def grid_shaped(self, grid_name, a):
      """Returns array reshaped to named grid extent for current use case, if known; otherwise unchanged."""

      grid_name = grid_name.upper()
      if self.use_case is not None and (self.use_case, grid_name) in self.grid_extents_kji.keys():
         cell_count = np.prod(self.grid_extents_kji[self.use_case, grid_name])
         if a.size == cell_count:
            return a.reshape(tuple(self.grid_extents_kji[self.use_case, grid_name]))
      return a


   def zip_glob(self, path_with_asterisk):
      """Performs glob.glob like function for zipped file, path must contain a single asterisk."""

      assert self.zipped

      star = path_with_asterisk.find('*')
      assert star >= 0, 'no asterisk in zip glob path'
      no_tail = (star == len(path_with_asterisk) - 1)
      match_list = []
      with zf.ZipFile(self.zip_file) as zfp:
         zip_list = zfp.namelist()
         for name in zip_list:
            if ((star == 0 or name.startswith(path_with_asterisk[:star])) and
                (no_tail or name.endswith(path_with_asterisk[star + 1:]))):
               match_list.append(name)
      return match_list



"""crs.py: RESQML coordinate reference system module."""

version = '2nd July 2021'

import logging
log = logging.getLogger(__name__)
log.debug('crs.py version ' + version)

import warnings
import math as maths
import numpy as np
# import xml.etree.ElementTree as et
# from lxml import etree as et

from resqpy.olio.base import BaseResqpy
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.olio.vector_utilities as vec
import resqpy.olio.weights_and_measures as wam
from resqpy.olio.xml_namespaces import curly_namespace as ns


class Crs(BaseResqpy):
   """ Coordinate reference system object """

   @property
   def resqml_type(self):
      return 'LocalTime3dCrs' if hasattr(self, 'time_units') and self.time_units else 'LocalDepth3dCrs'

   valid_axis_orders = ("easting northing", "northing easting", "westing southing",
                        "southing westing", "northing westing", "westing northing")

   def __init__(self, parent_model, crs_root = None, uuid = None,
                x_offset = 0.0, y_offset = 0.0, z_offset = 0.0,
                rotation = 0.0, xy_units = 'm', z_units = 'm',
                z_inc_down = True, axis_order = 'easting northing',
                time_units = None, epsg_code = None,
                title = None, originator = None, extra_metadata = None):
      """Create a new coordinate reference system object.

      arguments:
         parent_model (model.Model): the model to which the new Crs object will belong
         crs_root (xml root node): DEPRECATED – use uuid instead; the xml root node for an existing RESQML crs object
         crs_uuid (uuid.UUID): the uuid of an existing RESQML crs object in model, from which to instantiate the
            resqpy Crs object; if present, all the remaining arguments are ignored
         x_offset, y_offset, z_offset (floats, default zero): the local origin within an implicit parent crs
         rotation (float, default zero): a projected view rotation (in the xy plane), in radians, relative to an
            implicit parent crs
         xy_units (str, default 'm'): the units applicable to x & y values; must be one of the RESQML length uom strings
         z_units (str, default 'm'): the units applicable to z depth values; must be one of the RESQML length uom strings
         z_inc_down (boolean, default True): if True, increasing z values indicate greater depth (ie. in direction of
            gravity); if False, increasing z values indicate greater elevation
         axis_order (str, default 'easting northing'): the compass directions of the positive x and y axes respectively
         time_units (str, optional): if present, the time units of the z values (for seismic datasets), in which case the
            z_units argument is irrelevant
         epsg_code (str, optional): if present, the EPSG code of the implicit parent crs
         title (str, optional): the citation title to use for a new crs;
            ignored if uuid or crs_root is not None
         originator (str, optional): the name of the person creating the crs, defaults to login id;
            ignored if uuid or crs_root is not None
         extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the crs;
            ignored if uuid or crs_root is not None

      returns:
         a new resqpy Crs object

      notes:
         although resqpy does not have full support for identifying a parent crs, the modifiers such as a local origin
         may be used on the assumption that all crs objects refer to the same implicit parent crs;
         there are two equivalent RESQML classes and which one is generated depends on whether the time_units argument
         is used, when instantiating from values;
         it is strongly encourage to call the create_xml() method immediately after instantiation (unless the resqpy
         crs is a temporary object) as the uuid may be modified at that point to re-use any existing equivalent RESQML
         crs object in the model

      :meta common:
      """

      self.xy_units = xy_units
      self.z_units = z_units
      self.time_units = time_units   # if None, z values are depth; if not None, z values are time (from seismic)
      self.z_inc_down = z_inc_down
      self.x_offset = x_offset
      self.y_offset = y_offset
      self.z_offset = z_offset
      self.rotation = rotation    # radians
      self.axis_order = axis_order
      self.epsg_code = epsg_code
      # following are derived attributes, set below
      self.rotated = None
      self.rotation_matrix = None
      self.reverse_rotation_matrix = None

      super().__init__(model = parent_model, uuid = uuid, title = title, originator = originator,
                       extra_metadata = extra_metadata, root_node = crs_root)

      assert self.axis_order in self.valid_axis_orders, 'invalid CRS axis order: ' + str(axis_order)

      self.rotated = (not maths.isclose(self.rotation, 0.0, abs_tol = 1e-8) and
                      not maths.isclose(self.rotation, 2.0 * maths.pi, abs_tol = 1e-8))
      if self.rotated:
         self.rotation_matrix = vec.rotation_matrix_3d_axial(2, maths.degrees(self.rotation))
         self.reverse_rotation_matrix = vec.rotation_matrix_3d_axial(2, maths.degrees(-self.rotation))

      self.null_transform = (maths.isclose(self.x_offset, 0.0, abs_tol = 1e-8) and
                             maths.isclose(self.y_offset, 0.0, abs_tol = 1e-8) and
                             maths.isclose(self.z_offset, 0.0, abs_tol = 1e-8) and
                             not self.rotated)


   def _load_from_xml(self):
      root_node = self.root
      flavour = rqet.node_type(root_node)
      assert flavour in ['obj_LocalDepth3dCrs', 'obj_LocalTime3dCrs']
      if flavour == 'obj_LocalTime3dCrs':
         self.time_units = rqet.find_tag_text(root_node, 'TimeUom')
         assert self.time_units
      else:
         self.time_units = None
      self.xy_units = rqet.find_tag_text(root_node, 'ProjectedUom')
      self.axis_order = rqet.find_tag_text(root_node, 'ProjectedAxisOrder')
      self.z_units = rqet.find_tag_text(root_node, 'VerticalUom')
      self.z_inc_down = rqet.find_tag_bool(root_node, 'ZIncreasingDownward')
      self.x_offset = rqet.find_tag_float(root_node, 'XOffset')
      self.y_offset = rqet.find_tag_float(root_node, 'YOffset')
      self.z_offset = rqet.find_tag_float(root_node, 'ZOffset')
      self.rotation = rqet.find_tag_float(root_node, 'ArealRotation')  # todo: extract uom attribute from this node
      parent_xy_crs = rqet.find_tag(root_node, 'ProjectedCrs')
      if parent_xy_crs is not None and rqet.node_type(parent_xy_crs) == 'ProjectedCrsEpsgCode':
         self.epsg_code = rqet.find_tag_text(parent_xy_crs, 'EpsgCode')    # should be an integer?
      else:
         self.epsg_code = None


   def is_right_handed_xyz(self):
      """Returns True if the xyz axes are right handed; False if left handed."""

      return self.axis_order in ["northing easting", "southing westing", "westing northing"] == self.z_inc_down


   def global_to_local(self, xyz, global_z_inc_down = True):
      """Convert a single xyz point from the parent coordinate reference system to this one."""

      x, y, z = xyz
      if self.x_offset != 0.0: x -= self.x_offset
      if self.y_offset != 0.0: y -= self.y_offset
      if global_z_inc_down != self.z_inc_down: z = -z
      if self.z_offset != 0.0: z -= self.z_offset
      if self.rotated:
         (x, y, z) = vec.rotate_vector(self.rotation_matrix, np.array((x, y, z)))
      return (x, y, z)


   def global_to_local_array(self, xyz, global_z_inc_down = True):
      """Convert in situ a numpy array of xyz points from the parent coordinate reference system to this one."""

      if self.x_offset != 0.0: xyz[..., 0] -= self.x_offset
      if self.y_offset != 0.0: xyz[..., 1] -= self.y_offset
      if global_z_inc_down != self.z_inc_down:
         z = np.negative(xyz[..., 2])
         xyz[..., 2] = z
      if self.z_offset != 0.0: xyz[..., 2] -= self.z_offset
      if self.rotated:
         a = vec.rotate_array(self.rotation_matrix, xyz)
         xyz[:] = a


   def local_to_global(self, xyz, global_z_inc_down = True):
      """Convert a single xyz point from this coordinate reference system to the parent one."""

      if self.rotated:
         (x, y, z) = vec.rotate_vector(self.reverse_rotation_matrix, np.array(xyz))
      else:
         (x, y, z) = xyz
      if self.x_offset != 0.0: x += self.x_offset
      if self.y_offset != 0.0: y += self.y_offset
      if self.z_offset != 0.0: z += self.z_offset
      if global_z_inc_down != self.z_inc_down: z = -z
      return (x, y, z)


   def local_to_global_array(self, xyz, global_z_inc_down = True):
      """Convert in situ a numpy array of xyz points from this coordinate reference system to the parent one."""

      if self.rotated:
         a = vec.rotate_array(self.reverse_rotation_matrix, xyz)
         xyz[:] = a
      if self.x_offset != 0.0: xyz[..., 0] += self.x_offset
      if self.y_offset != 0.0: xyz[..., 1] += self.y_offset
      if self.z_offset != 0.0: xyz[..., 2] += self.z_offset
      if global_z_inc_down != self.z_inc_down:
         z = np.negative(xyz[..., 2])
         xyz[..., 2] = z


   def has_same_epsg_code(self, other_crs):
      """Returns True if either of the crs'es has a null EPSG code, or if they are the same."""
      return self.epsg_code is None or other_crs.epsg_code is None or self.epsg_code == other_crs.epsg_code


   def is_equivalent(self, other_crs):
      """Returns True if this crs is effectively the same as the other crs."""

      log.debug('testing crs equivalence')
      if other_crs is None: return False
      if self is other_crs: return True
      if bu.matching_uuids(self.uuid, other_crs.uuid): return True
      if self.xy_units != other_crs.xy_units or self.z_units != other_crs.z_units: return False
      if self.z_inc_down != other_crs.z_inc_down: return False
      if (self.time_units is not None or other_crs.time_units is not None) and self.time_units != other_crs.time_units:
         return False
      if not self.has_same_epsg_code(other_crs): return False
      if self.null_transform and other_crs.null_transform: return True
      if (maths.isclose(self.x_offset, other_crs.x_offset, abs_tol = 1e-4) and
          maths.isclose(self.y_offset, other_crs.y_offset, abs_tol = 1e-4) and
          maths.isclose(self.z_offset, other_crs.z_offset, abs_tol = 1e-4) and
          maths.isclose(self.rotation, other_crs.rotation, abs_tol = 1e-4)): return True
          # todo: handle and check rotation units; modularly equivalent rotations
      return False


   def convert_to(self, other_crs, xyz):
      """Converts a single xyz point from this coordinate reference system to the other.

      :meta common:
      """

      if self is other_crs: return tuple(xyz)
      assert self.has_same_epsg_code(other_crs)
      xyz = self.local_to_global(xyz)
      xyz = (wam.convert_lengths(xyz[0], self.xy_units, other_crs.xy_units),
             wam.convert_lengths(xyz[1], self.xy_units, other_crs.xy_units),
             wam.convert_lengths(xyz[2], self.z_units, other_crs.z_units))
      xyz = other_crs.global_to_local(xyz)
      return tuple(xyz)


   def convert_array_to(self, other_crs, xyz):
      """Converts in situ a numpy array of xyz points from this coordinate reference system to the other.

      :meta common:
      """

      if self.is_equivalent(other_crs): return
      assert self.has_same_epsg_code(other_crs)
      self.local_to_global_array(xyz)
      if self.xy_units == self.z_units and other_crs.xy_units == other_crs.z_units:
         wam.convert_lengths(xyz, self.xy_units, other_crs.xy_units)
      else:
         wam.convert_lengths(xyz[..., :2], self.xy_units, other_crs.xy_units)
         wam.convert_lengths(xyz[..., 2], self.z_units, other_crs.z_units)
      other_crs.global_to_local_array(xyz)
      return xyz


   def convert_from(self, other_crs, xyz):
      """Converts a single xyz point from the other coordinate reference system to this one.

      :meta common:
      """

      if self is other_crs: return tuple(xyz)
      assert self.has_same_epsg_code(other_crs)
      xyz = other_crs.local_to_global(xyz)
      xyz = (wam.convert_lengths(xyz[0], other_crs.xy_units, self.xy_units),
             wam.convert_lengths(xyz[1], other_crs.xy_units, self.xy_units),
             wam.convert_lengths(xyz[2], other_crs.z_units, self.z_units))
      xyz = self.global_to_local(xyz)
      return tuple(xyz)


   def convert_array_from(self, other_crs, xyz):
      """Converts in situ a numpy array of xyz points from the other coordinate reference system to this one.

      :meta common:
      """

      if self.is_equivalent(other_crs): return
      assert self.has_same_epsg_code(other_crs)
      other_crs.local_to_global_array(xyz)
      if self.xy_units == self.z_units and other_crs.xy_units == other_crs.z_units:
         wam.convert_lengths(xyz, other_crs.xy_units, self.xy_units)
      else:
         wam.convert_lengths(xyz[..., :2], other_crs.xy_units, self.xy_units)
         wam.convert_lengths(xyz[..., 2], other_crs.z_units, self.z_units)
      self.global_to_local_array(xyz)
      return xyz


   def create_xml(self, add_as_part = True, root = None, title = None, originator = None, reuse = True):
      """Creates a Coordinate Reference System xml node and optionally adds as a part in the parent model.

      arguments:
         add_as_part (boolean, default True): if True the newly created crs node is added to the model
            as a part
         root (optional, usually None): if not None, the newly created crs node is appended as a child
            of this node (rarely used)
         title (string, optional): used as the Title text in the citation node
         originator (string, optional): the name of the human being who created the crs object;
            default is to use the login name
         reuse (boolean, default True): if True and an equivalent crs already exists in the model then
            the uuid for this Crs is modified to match that of the existing object and the existing
            xml node is returned without anything new being added

      returns:
         newly created (or reused) coordinate reference system xml node

      notes:
         if the reuse argument is True, it is strongly recommended to call this method immediately after
         a new Crs has been instantiated from explicit values, as the uuid may be modified here;
         if reuse is True, the title is not regarded as important and a match with an existing object
         may occur even if the titles differ

      :meta common:
      """

      if reuse and self.try_reuse(): return self.node  # check for reusable (equivalent) object

      crs = super().create_xml(add_as_part = False, originator = originator)

      xoffset = rqet.SubElement(crs, ns['resqml2'] + 'XOffset')
      xoffset.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
      xoffset.text = '{0:5.3f}'.format(self.x_offset)

      yoffset = rqet.SubElement(crs, ns['resqml2'] + 'YOffset')
      yoffset.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
      yoffset.text = '{0:5.3f}'.format(self.y_offset)

      zoffset = rqet.SubElement(crs, ns['resqml2'] + 'ZOffset')
      zoffset.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
      zoffset.text = '{0:6.4f}'.format(self.z_offset)

      rotation = rqet.SubElement(crs, ns['resqml2'] + 'ArealRotation')
      rotation.set('uom', 'rad')
      rotation.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleMeasure')
      rotation.text = '{0:8.6f}'.format(self.rotation)

      if self.axis_order is None: axes = 'easting northing'
      else: axes = self.axis_order.lower()
      axis_order = rqet.SubElement(crs, ns['resqml2'] + 'ProjectedAxisOrder')
      axis_order.set(ns['xsi'] + 'type', ns['eml'] + 'AxisOrder2d')
      axis_order.text = axes

      xy_uom = rqet.SubElement(crs, ns['resqml2'] + 'ProjectedUom')
      xy_uom.set(ns['xsi'] + 'type', ns['eml'] + 'LengthUom')
      xy_uom.text = wam.rq_length_unit(self.xy_units)

      z_uom = rqet.SubElement(crs, ns['resqml2'] + 'VerticalUom')
      z_uom.set(ns['xsi'] + 'type', ns['eml'] + 'LengthUom')
      z_uom.text = wam.rq_length_unit(self.z_units)

      if self.time_units is not None:
         t_uom = rqet.SubElement(crs, ns['resqml2'] + 'TimeUom')
         t_uom.set(ns['xsi'] + 'type', ns['eml'] + 'TimeUom')  # todo: check this
         t_uom.text = self.time_units

      z_sense = rqet.SubElement(crs, ns['resqml2'] + 'ZIncreasingDownward')
      z_sense.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
      z_sense.text = str(self.z_inc_down).lower()

      xy_crs = rqet.SubElement(crs, ns['resqml2'] + 'ProjectedCrs')
      xy_crs.text = rqet.null_xml_text
      if self.epsg_code is None:
         xy_crs.set(ns['xsi'] + 'type', ns['eml'] + 'ProjectedUnknownCrs')
         self.model.create_unknown(root = xy_crs)
      else:
         xy_crs.set(ns['xsi'] + 'type', ns['eml'] + 'ProjectedCrsEpsgCode')
         epsg_node = rqet.SubElement(xy_crs, ns['resqml2'] + 'EpsgCode')
         epsg_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
         epsg_node.text = str(self.epsg_code)

      z_crs = rqet.SubElement(crs, ns['resqml2'] + 'VerticalCrs')
      if self.epsg_code is None:
         z_crs.set(ns['xsi'] + 'type', ns['eml'] + 'VerticalUnknownCrs')
         z_crs.text = rqet.null_xml_text
         self.model.create_unknown(root = z_crs)
      else:   # not sure if this is appropriate for the vertical crs
         z_crs.set(ns['xsi'] + 'type', ns['eml'] + 'VerticalCrsEpsgCode')
         epsg_node = rqet.SubElement(xy_crs, ns['resqml2'] + 'EpsgCode')
         epsg_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
         epsg_node.text = str(self.epsg_code)

      if root is not None: root.append(crs)
      if add_as_part: self.model.add_part('obj_' + self.resqml_type, bu.uuid_from_string(crs.attrib['uuid']), crs)
      if self.model.crs_root is None: self.model.crs_root = crs  # mark's as 'main' (ie. first) crs for model

      return crs


   @property
   def crs_root(self):
      """DEPRECATED. Alias for root"""
      warnings.warn("Attribute 'crs_root' is deprecated. Use 'root'", DeprecationWarning)
      return self.root



"""triangulation.py: functions for finding Delaunay triangulation and Voronoi graph from a set of points."""

version = '29th April 2021'

import numpy as np

import resqpy.olio.vector_utilities as vec


# _ccw_t() no longer needed: triangle vertices maintained in anti-clockwise order throughout
# def _ccw_t(p, t):   # puts triangle vertex indices into anti-clockwise order, in situ
#    if vec.clockwise(p[t[0]], p[t[1]], p[t[2]]) > 0.0:
#       t[1], t[2] = t[2], t[1]


def _dt_simple(po, plot_fn = None, progress_fn = None):

   def flip(ei):
      nonlocal fm, e, t, te, p, nt, p_i, ne
      if fm[ei]: return  # this edge has already been flipped since last point insertion
      t0, te0 = e[ei, 0]
      t1, te1 = e[ei, 1]
      if t0 < 0 or t1 < 0: return  # no triangle on other side
      t0n = te0 - 1
      if t0n < 0: t0n = 2
      t1n = te1 - 1
      if t1n < 0: t1n = 2
      if (not vec.in_circumcircle(p[t[t0, 0]], p[t[t0, 1]], p[t[t0, 2]], p[t[t1, t1n]]) and
          not vec.in_circumcircle(p[t[t1, 0]], p[t[t1, 1]], p[t[t1, 2]], p[t[t0, t0n]])): return  # not sure both are needed
      # flip needed
      ft0 = (t[t0, te0], t[t1, t1n], t[t0, t0n])
      ft1 = (t[t1, te1], t[t0, t0n], t[t1, t1n])
      e[ei, :, 1] = 1
      fte0 = (te[t1, 3 - (te1 + t1n)], ei, te[t0, t0n])
      fte1 = (te[t0, 3 - (te0 + t0n)], ei, te[t1, t1n])
      if e[fte0[0], 0, 0] == t1: e[fte0[0], 0] = (t0, 0)
      elif e[fte0[0], 1, 0] == t1: e[fte0[0], 1] = (t0, 0)
      else: raise Exception('edge breakdown')
      if e[fte1[0], 0, 0] == t0: e[fte1[0], 0] = (t1, 0)
      elif e[fte1[0], 1, 0] == t0: e[fte1[0], 1] = (t1, 0)
      else: raise Exception('edge breakdown')
      if e[fte0[2], 0, 0] == t0: e[fte0[2], 0, 1] = 2
      elif e[fte0[2], 1, 0] == t0: e[fte0[2], 1, 1] = 2
      else: raise Exception('edge breakdown')
      if e[fte1[2], 0, 0] == t1: e[fte1[2], 0, 1] = 2
      elif e[fte1[2], 1, 0] == t1: e[fte1[2], 1, 1] = 2
      else: raise Exception('edge breakdown')
      t[t0] = ft0
      t[t1] = ft1
      te[t0] = fte0
      te[t1] = fte1
      fm[ei] = True
      # debug plot here
      if plot_fn: plot_fn(p, t[:nt])
      # recursively flip, not sure all these are needed
      flip(fte0[0])
      flip(fte0[2])
      flip(fte1[0])
      flip(fte1[2])

   n_p = len(po)
   if n_p < 3: return None   # not enough points
   if progress_fn: progress_fn(0.0)

   min_xy = np.min(po[:, :2], axis = 0)
   max_xy = np.max(po[:, :2], axis = 0)
   dxy = max_xy - min_xy
   assert dxy[0] > 0.0 and dxy[1] > 0.0, 'points lie in straight line or are conincident'
   p = np.empty((n_p + 3, 2))
   p[:-3] = po[:, :2]
   # add 3 points sure of containing all po
   p[-3] = (min_xy[0] - 0.8 * dxy[0], min_xy[1] - 0.1 * dxy[1])
   p[-2] = (max_xy[0] + 0.8 * dxy[0], min_xy[1] - 0.1 * dxy[1])
   p[-1] = (min_xy[0] + 0.5 * dxy[0], max_xy[1] + 0.8 * dxy[1])

   # triangle vertex indices
   t = np.empty((2 * n_p + 2, 3), dtype = int)  # empty space for triangle vertex indices
   t[0] = (n_p, n_p + 1, n_p + 2)        # initial set of one contining triangle
   nt = 1                             # number of triangles so far populated

   # edges: list of indices of triangles and edge within triangle; -1 indicates no triangle using edge
   e = np.full((3 * n_p + 6, 2, 2), fill_value = -1, dtype = int)
   e[:3, 0, 0] = 0
   for edge in range(3):
      e[edge, 0, 1] = edge
   ne = 3                             # number of edges so far in use

   # edge indices (in e) for each triangle, first axis indexed in sync with t
   te = np.empty((2 * n_p + 2, 3), dtype = int)  # empty space for triangle edge indices
   te[0] = (0, 1, 2)

   # mask tracking which edges have been flipped
   fm = np.zeros((3 * n_p + 6), dtype = bool)

   # debug plot here
#   if plot_fn: plot_fn(p, t[:nt])

   progress_period = max(n_p // 100, 1)
   progress_count = progress_period

   for p_i in range(n_p):  # index of next point to consider

      if progress_fn and progress_count <= 0:
         progress_fn(float(p_i) / float(n_p))
         progress_count = progress_period

      # find triangle that contains this point
      f_t = None

      for ti in range(nt):
         if vec.in_triangle_edged(p[t[ti, 0]], p[t[ti, 1]], p[t[ti, 2]], p[p_i]):
            f_t = ti
            break
      assert f_t is not None, 'failed to find triangle containing point'

      # take copy of edge indices for containing triangle
      e0, e1, e2 = te[f_t]

      # split containing triangle into 3, using new point as common vertex
      t[nt] = (t[f_t, 1], t[f_t, 2], p_i)
      t[nt + 1] = (t[f_t, 2], t[f_t, 0], p_i)
      t[f_t, 2] = p_i
      # add 3 new edges and update te
      e[ne, 0] = (f_t, 2)
      e[ne, 1] = (nt + 1, 1)
      e[ne + 1, 0] = (f_t, 1)
      e[ne + 1, 1] = (nt, 2)
      e[ne + 2, 0] = (nt, 1)
      e[ne + 2, 1] = (nt + 1, 2)
      if e[e1, 0, 0] == f_t: e[e1, 0] = (nt, 0)
      elif e[e1, 1, 0] == f_t: e[e1, 1] = (nt, 0)
      else: raise Exception('edge breakdown')
      if e[e2, 0, 0] == f_t: e[e2, 0] = (nt + 1, 0)
      elif e[e2, 1, 0] == f_t: e[e2, 1] = (nt + 1, 0)
      else: raise Exception('edge breakdown')
      te[nt] = (e1, ne + 2, ne + 1)
      te[nt + 1] = (e2, ne, ne + 2)
      te[f_t, 1] = ne + 1
      te[f_t, 2] = ne

      nt += 2
      ne += 3

      # now recursively try flipping sides
      fm[:] = False

      # debug plot here, with new point, before flipping
      if plot_fn: plot_fn(p, t[:nt])

      flip(e0)
      flip(e1)
      flip(e2)

      progress_count -= 1

   # remove any triangles using invented container vertices
   tri_set = t[np.where(np.all(t < n_p, axis = 1))]
   if plot_fn: plot_fn(p, tri_set)
   if progress_fn: progress_fn(1.0)

   return tri_set


def dt(p, algorithm = None, plot_fn = None, progress_fn = None):
   """Returns the Delauney Triangulation of 2D point set p.

   arguments:
      p (numpy float array of shape (N, 2): the x,y coordinates of the points
      algorithm (string, optional): selects which algorithm to use; current options: ['simple'];
         if None, the current best algorithm is selected
      plot_fn (function of form f(p, t), optional): if present, this function is called each time the
         algorithm feels it is worth refreshing a plot of the progress; p is a copy of the point set,
         depending on the algorithm with 3 extra points added to form an enveloping triangle
      progress_fn (function of form f(x), optional): if present, this function is called at regulat
         intervals by the algorithm, passing increasing values in the range 0.0 to 1.0 as x

   returns:
      numpy int array of shape (M, 3) being the indices into the first axis of p of the 3 points
         per triangle in the Delauney Triangulation
   """
   global plot_out

   assert p.ndim == 2 and p.shape[1] >= 2, 'bad points shape for 2D Delauney Triangulation'

   if not algorithm: algorithm = 'simple'

   if algorithm == 'simple':
      return _dt_simple(p, plot_fn = plot_fn, progress_fn = progress_fn)
   else:
      raise Exception('unrecognised Delauney Triangulation algorithm name')



# factors.py
"""Module providing factorization and functions supporting grid extent determination from corner points."""

version = '24th December 2019'

def factorize(n):
   """Returns list of prime factors of positive integer n."""
   i = 2
   l = []
   while True:
      q, r = divmod(n, i)
      if r:
         i += 1
         if i > n: return l
      else:
         l.append(i)
         if q < i: return l
         n = q


def combinatorial(l):
   """Returns a list of all possible product combinations of numbers from list l, with some duplicates."""
   if len(l) == 0: return []
   head = l[0]
   c = [head]
   tail = combinatorial(l[1:])
   for o in tail:
      if o != head: c.append(o)
      c.append(head * o)
   return sorted(c)


def all_factors_from_primes(primes):
   """Returns a sorted list of unique factors from prime factorization."""
   all = list(set(combinatorial(primes)))
   all.append(1)
   return sorted(all)


def all_factors(n):
   """Returns a sorted list of unique factors of n."""
   primes = factorize(n)
   return all_factors_from_primes(primes)


def remove_subset(primary, subset):
   """Remove all elements of subset from primary list."""

   for e in subset:
      primary.remove(e)



"""read_nexus_fault.py: functions for reading Nexus fault definition data from an ascii file."""

version = '15th January 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging
log = logging.getLogger(__name__)

import os
import re
import unicodedata
import numpy as np
import pandas as pd


######################################################################################################
# load_nexus_fault_mult_table():
# function to load fault data for cell faces from nexus input deck format ascii file
# returns a pandas dataframe
#    compatible with that returned by load_table_from_file()
# note: actual multiplier value is discarded as this is usually set to one in the mult data
# this version written by Qiang

def load_nexus_fault_mult_table(file_name):

   """Reads a Nexus (!) format file containing one or more MULT keywords and returns a dataframe with the MULT rows."""

   def is_number(s):
      try:
         float(s)
         return True
      except ValueError:
         pass

      try:
         unicodedata.numeric(s)
         return True
      except (TypeError, ValueError):
         pass

      return False

   names = []
   grids = []
   faces = []
   dfs = []

   num_tables = 0
   ISTABLE = False
   ISRECORD = False
   grid = name = face = None

   face_dict = {'TX': 'I', 'TY': 'J', 'TZ': 'K', 'TI': 'I', 'TJ': 'J', 'TK': 'K'}

   if os.path.isfile(file_name):
      chunks = []
      with open(file_name) as f:
         for line in f:
            if len(line.strip()):
               if (not line.strip()[0] == '!') & (not line.strip()[0] == 'C') :
                  line = line.partition('!')[0]  # removing trailing comments
                  # line = line.partition('C')[0]  # removing trailing comments
                  tokens = line.split()
                  if ISTABLE:
                     if is_number(tokens[0]):
                        ISRECORD = True

                     if ISRECORD and (not is_number(tokens[0])):
                        data = chunks[0:]
                        d_elems = np.array([np.array(data[i].split()) for i in range(len(data))])
                        # fill empty elements with zero
                        lens = np.array([len(i) for i in d_elems])
                        # Mask of valid places in each row
                        mask = np.arange(lens.max()) < lens[:,None]
                        # Setup output array and put elements from data into masked positions
                        outdata = np.zeros(mask.shape, dtype=d_elems.dtype)
                        outdata[mask] = np.concatenate(d_elems)
                        df = pd.DataFrame(outdata)
                        for column in df.columns:
                           df[column] = pd.to_numeric(df[column], errors='ignore')
                        df.columns = ['i1','i2','j1','j2','k1','k2','mult']
                        grids.append(grid)
                        names.append(name)
                        faces.append(face)
                        dfs.append(df)
                        num_tables += 1

                        ISTABLE = False
                        ISRECORD = False
                        chunks = []

                  if ISTABLE:
                     if re.match("(.*)GRID(.*)", tokens[0]):
                        if len(tokens) > 0: grid = tokens[1]
                     elif re.match("(.*)FNAME(.*)", tokens[0]):
                        if len(tokens) > 0: name = tokens[1]
                     else:
                        if re.match(r"^MULT$", tokens[0]):
                           ISTABLE = False
                           ISRECORD = False
                           chunks = []
                        else:
                           chunks.append(line.strip())

                  if re.match(r"^MULT$", tokens[0]):
                     if len(tokens) > 0: face = face_dict[tokens[1]]
                     if 'MINUS' in tokens: face += '-'  # indicates data apply to 'negative' faces of specified cells
                     grid = 'ROOT'                    # nexus default
                     name = 'NONE'
                     ISTABLE = True

         else:
            if ISTABLE:
               if ISRECORD:
                  data = chunks[0:]
                  d_elems = np.array([np.array(data[i].split()) for i in range(len(data))])
                  # fill empty elements with zero
                  lens = np.array([len(i) for i in d_elems])
                  # Mask of valid places in each row
                  mask = np.arange(lens.max()) < lens[:,None]
                  # Setup output array and put elements from data into masked positions
                  outdata = np.zeros(mask.shape, dtype=d_elems.dtype)
                  outdata[mask] = np.concatenate(d_elems)
                  df = pd.DataFrame(outdata)
                  for column in df.columns:
                     df[column] = pd.to_numeric(df[column], errors='ignore')
                  df.columns = ['i1','i2','j1','j2','k1','k2','mult']

                  grids.append(grid)
                  names.append(name)
                  faces.append(face)
                  dfs.append(df)
                  num_tables += 1

                  ISTABLE = False
                  ISRECORD = False
                  chunks = []

   fault_df = pd.DataFrame()
   for i in range(0, len(dfs)):
      dfi = dfs[i]
      dfi['grid'] = grids[i]
      dfi['name'] = names[i]
      dfi['face'] = faces[i]
      fault_df = fault_df.append(dfi, ignore_index = True)

   convert_dict = {'i1': int, 'i2': int, 'j1': int, 'j2': int, 'k1': int, 'k2': int, 'mult': float}
   fault_df = fault_df.astype(convert_dict)

   return fault_df


def load_nexus_fault_mult_table_new(file_name):
   """Reads a Nexus (!) format file containing one or more MULT keywords and returns a dataframe with the MULT rows."""

   return load_nexus_fault_mult_table(file_name)



import pytest
from resqpy import model


def test_empty_model():
    _ = model.Model()
    return


def test_all_imports():
    from resqpy import crs, derived_model, fault, grid, grid_surface, lines, organize
    from resqpy import property, rq_import, surface, time_series, well
#    from resqpy.olio import *
    return



from pathlib import Path
from resqpy.organize import WellboreFeature

import pytest
import numpy as np
import pandas as pd
from numpy.testing import assert_array_almost_equal
from pandas.testing import assert_frame_equal

from resqpy.model import Model
from resqpy.crs import Crs
import resqpy.well


def test_MdDatum(example_model_and_crs):

   # Set up a new datum
   model, crs = example_model_and_crs
   epc = model.epc_file
   data = dict(
      location=(0, -99999, 3.14),
      md_reference='mean low water',
   )
   datum = resqpy.well.MdDatum(
      parent_model=model, crs_root=crs.root, **data
   )
   uuid = datum.uuid

   # Save to disk and reload
   datum.create_xml()
   model.store_epc()

   del model, crs, datum
   model2 = Model(epc_file=epc)
   datum2 = resqpy.well.MdDatum(parent_model=model2, uuid=uuid)
   
   for key, expected_value in data.items():
      assert getattr(datum2, key) == expected_value, f"Issue with {key}"


def test_trajectory_iterators(example_model_with_well):
   # Test that trajectories can be found via iterators

   model, well_interp, datum, traj = example_model_with_well

   # Iterate via wells
   uuids_1 = []
   for well in model.iter_wellbore_interpretations():
      for traj in well.iter_trajectories():
         uuids_1.append(traj.uuid)

   # Iterate directly
   uuids_2 = []
   for traj in model.iter_trajectories():
      uuids_2.append(traj.uuid)

   # Two sets of trajectories should match, assuming all trajectories have a parent well
   assert len(uuids_1) > 0
   assert sorted(uuids_1) == sorted(uuids_2)


def test_logs(example_model_with_logs):

   model, well_interp, datum, traj, frame, log_collection = example_model_with_logs

   # Check logs can be extracted from resqml dataset

   # Check exactly one wellbore frame exists in the model
   discovered_frames = 0
   for trajectory in model.iter_trajectories():
      for frame in trajectory.iter_wellbore_frames():
         discovered_frames += 1
   assert discovered_frames == 1

   # Check MDs
   mds = frame.node_mds
   assert isinstance(mds, np.ndarray)
   assert_array_almost_equal(mds, [1,2,3,4])

   # Check logs
   log_list = list(log_collection.iter_logs())
   assert len(log_list) == 2

   # TODO: would be nice to write: log_collection.get_curve("GR")
   gr = log_list[0]
   assert gr.title == "GR"
   assert gr.uom == "gAPI"
   assert_array_almost_equal(gr.values(), [1,2,1,2])

   nphi = log_list[1]
   assert nphi.title == 'NPHI'

   # TODO: get more units working
   # assert nphi.uom == "v/v"
   assert_array_almost_equal(nphi.values(), [0.1, 0.1, np.NaN, np.NaN])
   

def test_logs_conversion(example_model_with_logs):

   model, well_interp, datum, traj, frame, log_collection = example_model_with_logs

   # Pandas
   df = log_collection.to_df()
   df_expected = pd.DataFrame(
      data={"GR": [1,2,1,2], "NPHI": [0.1, 0.1, np.NaN, np.NaN]},
      index=[1,2,3,4]
   )
   assert_frame_equal(df_expected, df, check_dtype=False)
   
   # LAS
   las = log_collection.to_las()
   assert las.well.WELL.value == 'well A'

   gr = las.get_curve("GR")
   assert gr.unit.casefold() == "GAPI".casefold()
   assert_array_almost_equal(gr.data, [1,2,1,2])

   nphi = las.get_curve("NPHI")
   # assert nphi.unit == "GAPI"
   assert_array_almost_equal(nphi.data, [0.1, 0.1, np.NaN, np.NaN])


# Trajectory

def test_Trajectory_add_well_feature_and_interp(example_model_and_crs):

   # Prepare an example Trajectory without a well feature
   wellname = "Hullabaloo"
   model, crs = example_model_and_crs
   datum = resqpy.well.MdDatum(
      parent_model=model, crs_root=crs.root, location=(0, 0, -100), md_reference='kelly bushing'
   )
   datum.create_xml()
   traj = resqpy.well.Trajectory(parent_model=model, md_datum=datum, well_name=wellname)


   # Add the well interp
   assert traj.wellbore_feature is None
   assert traj.wellbore_interpretation is None
   traj.create_feature_and_interpretation()

   # Check well is present
   assert traj.wellbore_feature is not None
   assert traj.wellbore_feature.feature_name == wellname


# Deviation Survey tests

def test_DeviationSurvey(example_model_with_well):
   # Test that all attrbutes are correctly saved and loaded from disk

   # --------- Arrange ----------
   # Create a Deviation Survey object in memory

   # Load example model from a fixture
   model, well_interp, datum, traj = example_model_with_well
   epc_path = model.epc_file

   # Create the survey
   data = dict(
      title='Majestic Umlaut ö',
      originator='Thor, god of sparkles',
      md_uom='ft',
      angle_uom='rad',
      is_final=True,
   )
   array_data = dict(
      measured_depths=np.array([1, 2, 3], dtype=float),
      azimuths=np.array([4, 5, 6], dtype=float),
      inclinations=np.array([7, 8, 9], dtype=float),
      first_station=np.array([0, -1, 999], dtype=float),
   )
   survey = resqpy.well.DeviationSurvey(
      parent_model=model,
      represented_interp=well_interp,
      md_datum=datum,
      **data,
      **array_data,
   )
   survey_uuid = survey.uuid

   # ----------- Act ---------

   # Save to disk
   survey.write_hdf5()
   survey.create_xml()
   model.store_epc()
   model.h5_release()

   # import shutil
   # shutil.copy(epc_path, r'C:\Temp')
   # shutil.copy(epc_path.replace('.epc', '.h5'), r'C:\Temp')

   # Clear memory
   del model, well_interp, datum, traj, survey

   # Reload from disk
   model2 = Model(epc_file=epc_path)
   survey2 = resqpy.well.DeviationSurvey(model2, uuid=survey_uuid)

   # --------- Assert --------------
   # Check all attributes were loaded from disk correctly

   for key, expected_value in data.items():
      assert getattr(survey2, key) == expected_value, f"Error for {key}"
   
   for key, expected_value in array_data.items():
      assert_array_almost_equal(
         getattr(survey2, key), expected_value, err_msg=f"Error for {key}"
      )
